# -*- coding: utf-8 -*-
"""Ticket related functions and classes are situated here."""
import uuid
from typing import Any, Dict, List, Optional, Union

from sqlalchemy import Column, Integer, String, Text
from sqlalchemy.exc import OperationalError, UnboundExecutionError
from sqlalchemy.orm import Mapped, mapped_column, relationship, synonym
from sqlalchemy.orm.mapper import validates
from sqlalchemy.schema import ForeignKey, Table
from sqlalchemy.types import Enum

from stalker.db.declarative import Base
from stalker.db.session import DBSession
from stalker.exceptions import CircularDependencyError
from stalker.log import get_logger
from stalker.models.auth import User
from stalker.models.entity import Entity, SimpleEntity
from stalker.models.mixins import StatusMixin
from stalker.models.note import Note
from stalker.models.project import Project
from stalker.models.status import Status

logger = get_logger(__name__)

# RESOLUTIONS
FIXED = "fixed"
INVALID = "invalid"
WONTFIX = "wontfix"
DUPLICATE = "duplicate"
WORKSFORME = "worksforme"
CANTFIX = "cantfix"


class Ticket(Entity, StatusMixin):
    """Tickets are the way of reporting errors or asking for changes.

    The Stalker Ticketing system is based on Trac Basic Workflow. For more
    information please visit `Trac Workflow`_

    _`Trac Workflow`:: http://trac.edgewall.org/wiki/TracWorkflow

    Stalker Ticket system is very flexible, to customize the workflow please
    update the :class:`.Config.ticket_workflow` dictionary.

    In the default setup, there are four actions available; ``accept``,
    ``resolve``, ``reopen``, ``reassign``, and five statuses available ``New``,
    ``Assigned``, ``Accepted``, ``Reopened``, ``Closed``.

    Args:
        project (Project): The Project that this Ticket is assigned to. A Ticket in
            Stalker must be assigned to a Project. ``project`` argument cannot be
            skipped or cannot be None.
        summary (str): A string which contains the title or a short
            description of this Ticket.
        priority (str): The priority of the Ticket which is an enum value.
            Possible values are:

            +--------------+-------------------------------------------------+
            | 0 / TRIVIAL  | defect with little or no impact / cosmetic      |
            |              | enhancement                                     |
            +--------------+-------------------------------------------------+
            | 1 / MINOR    | defect with minor impact / small enhancement    |
            +--------------+-------------------------------------------------+
            | 2 / MAJOR    | defect with major impact / big enhancement      |
            +--------------+-------------------------------------------------+
            | 3 / CRITICAL | severe loss of data due to the defect or highly |
            |              | needed enhancement                              |
            +--------------+-------------------------------------------------+
            | 4 / BLOCKER  | basic functionality is not available until this |
            |              | is fixed                                        |
            +--------------+-------------------------------------------------+

        reported_by (User): A :class:`.User` instance who created this Ticket. It is
            basically a synonym for the :attr:`.SimpleEntity.created_by` attribute.

    Changing the :attr`.Ticket.status` will create a new :class:`.TicketLog` instance
    showing the previous operation.

    Even though Tickets needs statuses they don't need to be supplied a
    :class:`.StatusList` nor :class:`.Status` for the Tickets. It will be
    automatically filled accordingly. For newly created Tickets the status of
    the ticket is ``NEW`` and can be changed to other statuses as follows:

        Status   -> Action   -> New Status

        NEW      -> resolve  -> CLOSED
        NEW      -> accept   -> ACCEPTED
        NEW      -> reassign -> ASSIGNED

        ASSIGNED -> resolve  -> CLOSED
        ASSIGNED -> accept   -> ACCEPTED
        ASSIGNED -> reassign -> ASSIGNED

        ACCEPTED -> resolve  -> CLOSED
        ACCEPTED -> accept   -> ACCEPTED
        ACCEPTED -> reassign -> ASSIGNED

        REOPENED -> resolve  -> CLOSED
        REOPENED -> accept   -> ACCEPTED
        REOPENED -> reassign -> ASSIGNED

        CLOSED   -> reopen   -> REOPENED

        actions available:
        resolve
        reassign
        accept
        reopen

    The :attr:`.Ticket.name` is automatically generated by using the
    ``stalker.config.Config.ticket_label`` attribute and
    :attr:`.Ticket.ticket_number` . So if defaults are used the first ticket
    name will be "Ticket#1" and the second "Ticket#2" and so on. For every
    project the number will restart from 1.

    Use the :meth:`.Ticket.resolve`, :meth:`.Ticket.reassign`,
    :meth:`.Ticket.accept`, :meth:`.Ticket.reopen` methods to change the status
    of the current Ticket.

    Changing the status of the Ticket will create :class:`.TicketLog` entries
    reflecting the change made.
    """

    # logs attribute
    __auto_name__ = True
    __tablename__ = "Tickets"
    # __table_args__ = (
    #    UniqueConstraint("project_id", 'number'), {}
    # )
    __mapper_args__ = {"polymorphic_identity": "Ticket"}

    ticket_id: Mapped[int] = mapped_column(
        "id", ForeignKey("Entities.id"), primary_key=True
    )

    # TODO: use ProjectMixin
    project_id: Mapped[int] = mapped_column("project_id", ForeignKey("Projects.id"))

    _project: Mapped[Project] = relationship(
        primaryjoin="Tickets.c.project_id==Projects.c.id",
        back_populates="tickets",
    )

    _number: Mapped[int] = mapped_column(
        "number",
        autoincrement=True,
        default=1,
        nullable=False,
        unique=True,
    )

    related_tickets: Mapped[Optional[List["Ticket"]]] = relationship(
        secondary="Ticket_Related_Tickets",
        primaryjoin="Tickets.c.id==Ticket_Related_Tickets.c.ticket_id",
        secondaryjoin="Ticket_Related_Tickets.c.related_ticket_id==" "Tickets.c.id",
        doc="""A list of other Ticket instances which are related
        to this one. Can be used to related Tickets to point to a common
        problem. The Ticket itself cannot be assigned to this list
        """,
    )

    summary: Mapped[Optional[str]] = mapped_column(Text)

    logs: Mapped[Optional[List["TicketLog"]]] = relationship(
        primaryjoin="Tickets.c.id==TicketLogs.c.ticket_id",
        back_populates="ticket",
        cascade="all, delete-orphan",
    )

    links: Mapped[Optional[List["SimpleEntity"]]] = relationship(
        secondary="Ticket_SimpleEntities"
    )

    comments: Mapped[Optional[List["Note"]]] = synonym(
        "notes",
        doc="""A list of :class:`.Note` instances showing the comments made for
        this Ticket instance. It is a synonym for the :attr:`.Ticket.notes`
        attribute.
        """,
    )

    reported_by: Mapped[Optional["User"]] = synonym(
        "created_by", doc="Shows who created this Ticket"
    )

    owner_id: Mapped[Optional[int]] = mapped_column(ForeignKey("Users.id"))
    owner: Mapped["User"] = relationship(primaryjoin="Tickets.c.owner_id==Users.c.id")

    resolution: Mapped[Optional[str]] = mapped_column(String(128))

    priority: Mapped[Optional[str]] = mapped_column(
        Enum("TRIVIAL", "MINOR", "MAJOR", "CRITICAL", "BLOCKER", name="PriorityType"),
        default="TRIVIAL",
        doc="""The priority of the Ticket which is an enum value.
        Possible values are:

          +--------------+-------------------------------------------------+
          | 0 / TRIVIAL  | defect with little or no impact / cosmetic      |
          |              | enhancement                                     |
          +--------------+-------------------------------------------------+
          | 1 / MINOR    | defect with minor impact / small enhancement    |
          +--------------+-------------------------------------------------+
          | 2 / MAJOR    | defect with major impact / big enhancement      |
          +--------------+-------------------------------------------------+
          | 3 / CRITICAL | severe loss of data due to the defect or highly |
          |              | needed enhancement                              |
          +--------------+-------------------------------------------------+
          | 4 / BLOCKER  | basic functionality is not available until this |
          |              | is fixed                                        |
          +--------------+-------------------------------------------------+
        """,
    )

    def __init__(
        self,
        project: Optional[Project] = None,
        links: Optional[List[SimpleEntity]] = None,
        priority: str = "TRIVIAL",
        summary: Optional[str] = None,
        **kwargs: Dict[str, Any],
    ) -> None:
        # just force auto name generation
        self._number = self._generate_ticket_number()
        from stalker import defaults

        kwargs["name"] = "{:s} #{:d}".format(defaults.ticket_label, self.number)

        super(Ticket, self).__init__(**kwargs)
        StatusMixin.__init__(self, **kwargs)

        self._project = project

        self.priority = priority
        if links is None:
            links = []
        self.links = links
        self.summary = summary

    def _number_getter(self) -> int:
        """Return the number attribute value.

        Returns:
            int: The number attribute value.
        """
        return self._number

    number = synonym(
        "_number",
        descriptor=property(_number_getter),
        doc="""The automatically generated number for the tickets.
        """,
    )

    def _project_getter(self) -> Project:
        """Return the project attribute value.

        Returns:
            Project: The project attribute value.
        """
        return self._project

    project = synonym("_project", descriptor=property(_project_getter))

    @classmethod
    def _maximum_number(cls) -> int:
        """Return the maximum available number from the database.

        Returns:
            int: The maximum ticket number.
        """
        try:
            # do your query
            with DBSession.no_autoflush:
                max_ticket = Ticket.query.order_by(Ticket.number.desc()).first()
        except (UnboundExecutionError, OperationalError):
            max_ticket = None

        return max_ticket.number if max_ticket is not None else 0

    def _generate_ticket_number(self) -> int:
        """Auto generate a number for the ticket.

        Returns:
            int: The auto generated ticket number.
        """
        # TODO: try to make it atomic
        return self._maximum_number() + 1

    @validates("related_tickets")
    def _validate_related_tickets(self, key: str, related_ticket: "Ticket") -> "Ticket":
        """Validate the given related_ticket value.

        Args:
            key (str): The name of the validated column.
            related_ticket (Ticket): The related_ticket value to be validated.

        Raises:
            TypeError: If the related_ticket value is not a Ticket instance.
            CircularDependencyError: If the related_ticket value is the same with this
                Ticket.

        Returns:
            Ticket: The validated related_ticket value.
        """
        if not isinstance(related_ticket, Ticket):
            raise TypeError(
                "{}.related_ticket should only contain instances of "
                "stalker.models.ticket.Ticket, not {}: '{}'".format(
                    self.__class__.__name__,
                    related_ticket.__class__.__name__,
                    related_ticket,
                )
            )

        if related_ticket is self:
            raise CircularDependencyError(
                "{}.related_ticket attribute cannot have itself in the list".format(
                    self.__class__.__name__
                )
            )

        return related_ticket

    @validates("_project")
    def _validate_project(
        self, key: str, project: Union[None, Project]
    ) -> Union[None, Project]:
        """Validate the given project value.

        Args:
            key (str): The name of the validated column.
            project (Union[None, Project]): The project value to be validated.

        Raises:
            TypeError: If the given project is not a Project instance.

        Returns:
            Union[None, Project]: The validated project value.
        """
        if project is None or not isinstance(project, Project):
            raise TypeError(
                "{}.project should be an instance of "
                "stalker.models.project.Project, not {}: '{}'".format(
                    self.__class__.__name__, project.__class__.__name__, project
                )
            )

        return project

    @validates("summary")
    def _validate_summary(self, key: str, summary: Union[None, str]) -> str:
        """Validate the given summary value.

        Args:
            key (str): The name of the validated column.
            summary (Union[None, str]): The summary value to be validated.

        Raises:
            TypeError: If the given summary is not None and not a string.

        Returns:
            str: The validated summary value.
        """
        if summary is None:
            summary = ""

        if not isinstance(summary, str):
            raise TypeError(
                "{}.summary should be an instance of str, not {}: '{}'".format(
                    self.__class__.__name__, summary.__class__.__name__, summary
                )
            )
        return summary

    def __action__(
        self, action: str, created_by: User, action_arg: Any = None
    ) -> "TicketLog":
        """Update the ticket status and create a ticket log.

        The log is created according to the Ticket.__available_actions__ dictionary.

        Args:
            action (str): The name of the action.
            created_by (User): The User creating this action.
            action_arg (Any): The argument to pass to the action.

        Returns:
            TicketLog: The TicketLog instance created.
        """
        from stalker import defaults

        statuses = defaults.ticket_workflow[action].keys()
        status = self.status.name
        return_value = None
        if status in statuses:
            action_data = defaults.ticket_workflow[action][status]
            new_status_code = action_data["new_status"]
            action_name = action_data["action"]

            # there is an action defined for this status
            # get the to_status
            from_status = self.status
            to_status = self.status_list[new_status_code]
            self.status = to_status

            # call the action with action_arg
            func = getattr(self, action_name)
            func(action_arg)

            ticket_log = TicketLog(
                self, from_status, to_status, action, created_by=created_by
            )

            # create log entry
            self.logs.append(ticket_log)
            return_value = ticket_log
        return return_value

    def resolve(
        self, created_by: Union[None, User] = None, resolution: str = ""
    ) -> "TicketLog":
        """Resolve the ticket.

        Args:
            created_by (User): The User instance who is taking this action.
            resolution (str): The resolution.

        Returns:
            TicketLog: The newly created TicketLog instance.
        """
        return self.__action__("resolve", created_by, resolution)

    def accept(self, created_by: Union[None, User] = None) -> "TicketLog":
        """Accept the ticket.

        Args:
            created_by (User): The User instance who is taking this action.

        Returns:
            TicketLog: The newly created TicketLog instance.
        """
        return self.__action__("accept", created_by, created_by)

    def reassign(
        self, created_by: Union[None, User] = None, assign_to: Union[None, User] = None
    ) -> "TicketLog":
        """Reassign the ticket to another User.

        Args:
            created_by (User): The User that is doing the action.
            assign_to (User): The new owner of the ticket.

        Returns:
            TicketLog: The newly created TicketLog instance.
        """
        return self.__action__("reassign", created_by, assign_to)

    def reopen(self, created_by: Union[None, User] = None) -> "TicketLog":
        """Reopen the ticket.

        Args:
            created_by (User): The User who is doing the action.

        Returns:
            TicketLog: The newly created TicketLog instance.
        """
        return self.__action__("reopen", created_by)

    # actions
    def set_owner(self, *args) -> None:
        """Set the owner.

        Args:
            args (Any): Set the owner.
        """
        self.owner = args[0]

    def set_resolution(self, *args) -> None:
        """Set the timing_resolution.

        Args:
            args (Any): Any argument passed to this method.
        """
        self.resolution = args[0]

    def del_resolution(self, *args) -> None:
        """Delete the timing_resolution.

        Args:
            args (Any): Any arguments passed to this method.
        """
        self.resolution = ""

    def __eq__(self, other: Any) -> bool:
        """Check the equality.

        Args:
            other (Any): The other object.

        Returns:
            bool: True if the other object is a Ticket instance and has the same name,
                number, status, logs and priority.
        """
        return (
            super(Ticket, self).__eq__(other)
            and isinstance(other, Ticket)
            and other.name == self.name
            and other.number == self.number
            and other.status == self.status
            and other.logs == self.logs
            and other.priority == self.priority
        )

    def __hash__(self) -> int:
        """Return the hash value of this instance.

        Because the __eq__ is overridden the __hash__ also needs to be overridden.

        Returns:
            int: The hash value.
        """
        return super(Ticket, self).__hash__()


class TicketLog(SimpleEntity):
    """Holds :attr:`.Ticket.status` change operations.

    Args:
        ticket (Ticket): A :class:`.Ticket` instance which is the subject of the
            operation.

        from_status (Status): Holds a reference to a :class:`.Status` instance which is
            the previous status of the :class:`.Ticket` .

        to_status (Status): Holds a reference to a :class:`.Status` instance which is
            the new status of the :class;`.Ticket` .

        action (str): An Enumerator holding the type of the operation should be one of
            ["resolve", "accept", "reassign", "reopen"].

      Operations follow the `Track Workflow`_ ,

      .. image:: http://trac.edgewall.org/chrome/common/guide/original-workflow.png
          :width: 787 px
          :height: 509 px
          :align: left

    .. _Track Workflow: http://trac.edgewall.org/wiki/TracWorkflow
    """

    from stalker import defaults  # need to limit it with a scope

    # TODO: there are no tests for the TicketLog class

    __tablename__ = "TicketLogs"
    __mapper_args__ = {"polymorphic_identity": "TicketLog"}

    ticket_log_id: Mapped[int] = mapped_column(
        "id", ForeignKey("SimpleEntities.id"), primary_key=True
    )

    from_status_id: Mapped[Optional[int]] = mapped_column(ForeignKey("Statuses.id"))
    to_status_id: Mapped[Optional[int]] = mapped_column(ForeignKey("Statuses.id"))

    from_status: Mapped[Status] = relationship(
        primaryjoin="TicketLogs.c.from_status_id==Statuses.c.id"
    )

    to_status: Mapped[Status] = relationship(
        primaryjoin="TicketLogs.c.to_status_id==Statuses.c.id"
    )

    action: Mapped[Optional[str]] = mapped_column(
        Enum(*defaults.ticket_workflow.keys(), name="TicketActions")
    )

    ticket_id: Mapped[Optional[int]] = mapped_column(ForeignKey("Tickets.id"))
    ticket: Mapped[Optional[Ticket]] = relationship(
        primaryjoin="TicketLogs.c.ticket_id==Tickets.c.id",
        back_populates="logs",
    )

    def __init__(
        self,
        ticket: Optional[Ticket] = None,
        from_status: Optional[Status] = None,
        to_status: Optional[Status] = None,
        action: Optional[str] = None,
        **kwargs: Dict[str, Any],
    ) -> None:
        kwargs["name"] = "TicketLog_" + uuid.uuid4().hex
        super(TicketLog, self).__init__(**kwargs)
        self.ticket = ticket
        self.from_status = from_status
        self.to_status = to_status
        self.action = action


# A secondary Table for Ticket to Ticket relations
Ticket_Related_Tickets = Table(
    "Ticket_Related_Tickets",
    Base.metadata,
    Column("ticket_id", Integer, ForeignKey("Tickets.id"), primary_key=True),
    Column("related_ticket_id", Integer, ForeignKey("Tickets.id"), primary_key=True),
    extend_existing=True,
)

# Ticket SimpleEntity Relation, link anything to a ticket
Ticket_SimpleEntities = Table(
    "Ticket_SimpleEntities",
    Base.metadata,
    Column("ticket_id", Integer, ForeignKey("Tickets.id"), primary_key=True),
    Column(
        "simple_entity_id", Integer, ForeignKey("SimpleEntities.id"), primary_key=True
    ),
)



# -*- coding: utf-8 -*-
"""Structure related functions and classes are situated here."""
from typing import Any, Dict, List, Optional, Union

from sqlalchemy import Column, ForeignKey, Integer, Table, Text
from sqlalchemy.orm import Mapped, mapped_column, relationship, validates

from stalker.db.declarative import Base
from stalker.log import get_logger
from stalker.models.entity import Entity
from stalker.models.template import FilenameTemplate

logger = get_logger(__name__)


class Structure(Entity):
    """Defines folder structures for :class:`.Projects`.

    Structures are generally owned by :class:`.Project` objects. Whenever a
    :class:`.Project` is physically created, project folders are created by
    looking at :attr:`.Structure.custom_template` of the :class:`.Structure`,
    the :class:`.Project` object is generally given to the :class:`.Structure`.
    So it is possible to use a variable like "{{project}}" or derived variables
    like::

      {% for seq in project.sequences %}
          {{do something here}}

    Every line of this rendered template will represent a folder and Stalker
    will create these folders on the attached :class:`.Repository`.

    Args:
        templates (List[FilenameTemplate]): A list of :class:`.FilenameTemplate`
            instances which defines a specific template for the given
            :attr:`.FilenameTemplate.target_entity_type` values.

        custom_template (str): A string containing several lines of folder
            names. The folders are relative to the :class:`.Project` root. It
            can also contain a Jinja2 Template code. Which will be rendered to
            show the list of folders to be created with the project. The Jinja2
            Template is going to have the {{project}} variable. The important
            point to be careful about is to list all the custom folders of the
            project in a new line in this string. For example a
            :class:`.Structure` for a :class:`.Project` can have the following
            :attr:`.Structure.custom_template`::

            .. code-block:: Jinja

                ASSETS
                {% for asset in project.assets %}
                    {% set asset_root = 'ASSETS/' + asset.code %}
                    {{asset_root}}

                    {% for task in asset.tasks %}
                        {% set task_root = asset_root + '/' + task.code %}
                        {{task_root}}

                SEQUENCES
                {% for seq in project.sequences %}}
                    {% set seq_root = 'SEQUENCES/' + {{seq.code}} %}
                    {{seq_root}}/Edit
                    {{seq_root}}/Edit/Export
                    {{seq_root}}/Storyboard

                    {% for shot in seq.shots %}
                        {% set shot_root = seq_root + '/SHOTS/' + shot.code %}
                        {{shot_root}}

                        {% for task in shot.tasks %}
                            {% set task_root = shot_root + '/' + task.code %}
                            {{task_root}}

            The above example has gone far beyond deep than it is needed, where
            it started to define paths for :class:`.Asset` s. Even it is
            possible to create a :class:`.Project` structure like that, in
            general it is unnecessary. Because the above folders are going to
            be created but they are probably going to be empty for a while,
            because the :class:`.Asset` s are not created yet (or in fact no
            :class:`.Version` instances are created for the :class:`.Task` s).
            Anyway, it is much suitable and desired to create this details by
            using :class:`.FilenameTemplate` objects. Which are specific to
            certain :attr:`.FilenameTemplate.target_entity_type` s. And by
            using the :attr:`.Structure.custom_template` attribute, Stalker
            cannot place any source or output file of a :class:`.Version` in
            the :class:`.Repository` where as it can by using
            :class:`.FilenameTemplate` s.

            But for certain types of :class:`.Task` s it is may be good to
            previously create the folder structure just because in certain
            environments (programs) it is not possible to run a Python code
            that will place the file in to the Repository like in Photoshop.

            The ``custom_template`` parameter can be None or an empty string if
            it is not needed.

            A :class:`.Structure` cannot be created without a ``type``
            (__strictly_typed__ = True). By giving a ``type`` to the
            :class:`.Structure`, you can create one structure for
            **Commercials** and another project structure for **Movies** and
            another one for **Print** projects etc. and can reuse them with new
            :class:`.Project` s.
    """

    # __strictly_typed__ = True
    __auto_name__ = False
    __tablename__ = "Structures"
    __mapper_args__ = {"polymorphic_identity": "Structure"}

    structure_id: Mapped[int] = mapped_column(
        "id",
        ForeignKey("Entities.id"),
        primary_key=True,
    )

    templates: Mapped[Optional[List[FilenameTemplate]]] = relationship(
        secondary="Structure_FilenameTemplates"
    )

    custom_template: Mapped[Optional[str]] = mapped_column("custom_template", Text)

    def __init__(
        self,
        templates: Optional[List[FilenameTemplate]] = None,
        custom_template: Optional[str] = None,
        **kwargs: Dict[str, Any],
    ) -> None:
        super(Structure, self).__init__(**kwargs)

        if templates is None:
            templates = []

        self.templates = templates
        self.custom_template = custom_template

    def __eq__(self, other: Any) -> bool:
        """Check the equality.

        Args:
            other (Any): The other object.

        Returns:
            bool: True if the other object is a Structure instance and has the same
                templates, custom_template.
        """
        return (
            super(Structure, self).__eq__(other)
            and isinstance(other, Structure)
            and self.templates == other.templates
            and self.custom_template == other.custom_template
        )

    def __hash__(self) -> int:
        """Return the hash value of this instance.

        Because the __eq__ is overridden the __hash__ also needs to be overridden.

        Returns:
            int: The hash value.
        """
        return super(Structure, self).__hash__()

    @validates("custom_template")
    def _validate_custom_template(
        self, key: str, custom_template: Union[None, str]
    ) -> str:
        """Validate the given custom_template value.

        Args:
            key (str): The name of the validated column.
            custom_template (Union[None, str]): The custom template value to be
                validated.

        Raises:
            TypeError: If the given custom_template value is not a str.

        Returns:
            str: The validated `custom_template` value.
        """
        if custom_template is None:
            custom_template = ""

        if not isinstance(custom_template, str):
            raise TypeError(
                "{}.custom_template should be a string, not {}: '{}'".format(
                    self.__class__.__name__,
                    custom_template.__class__.__name__,
                    custom_template,
                )
            )
        return custom_template

    @validates("templates")
    def _validate_templates(
        self, key: str, template: FilenameTemplate
    ) -> FilenameTemplate:
        """Validate the given template value.

        Args:
            key (str): The name of the validated column.
            template (FilenameTemplate): The validated template value.

        Raises:
            TypeError: If the given template value is not a FilenameTemplate
                instance.

        Returns:
            FilenameTemplate: Return the validated template value.
        """
        if not isinstance(template, FilenameTemplate):
            raise TypeError(
                f"{self.__class__.__name__}.templates should only contain "
                "instances of stalker.models.template.FilenameTemplate, "
                f"not {template.__class__.__name__}: '{template}'"
            )

        return template


# Structure_FilenameTemplates Table
Structure_FilenameTemplates = Table(
    "Structure_FilenameTemplates",
    Base.metadata,
    Column("structure_id", Integer, ForeignKey("Structures.id"), primary_key=True),
    Column(
        "filenametemplate_id",
        Integer,
        ForeignKey("FilenameTemplates.id"),
        primary_key=True,
    ),
)



# -*- coding: utf-8 -*-
"""Client related classes and functions are situated here."""

from typing import Any, Dict, List, Optional, TYPE_CHECKING

from sqlalchemy import ForeignKey
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.orm import Mapped, mapped_column, relationship, validates

from stalker import log
from stalker.db.declarative import Base
from stalker.models.entity import Entity
from stalker.models.project import create_project_client

if TYPE_CHECKING:  # pragma: no cover
    from stalker.models.auth import Role, User
    from stalker.models.budget import Good
    from stalker.models.project import Project, ProjectClient

logger = log.get_logger(__name__)


class Client(Entity):
    """The Client (e.g. a company) which users may be part of.

    The information that a Client object holds is like:

      * The users of the client
      * The projects affiliated with the client
      * and all the other things those are inherited from the Entity class

    .. note::
       .. versionadded 0.2.20: Client Specific Goods

       Clients now can own a list of :class:`.Good` s attached to them.
       So one can define a list of class:`.Good` s with special prices
       adjusted for a particular ``Client``, then get them back from the db by
       querying the :class:`.Good` s those have their ``client`` attribute set
       to that particular ``Client`` instance. Removing a ``Good`` from a
       :class:`.Client` will not delete it from the database, but deleting a
       :class:`.Client` will also delete the ``Good`` s attached to that
       particular :class:`.Client`.

    .. ::
       don't forget to update the Good documentation, which also has the same
       text.

    Two Client object considered the same if they have the same name.

    So creating a client object needs the following parameters:

    Args:
        users (:class:`.User`): It can be an empty list, so one client can be created
            without any user in it. But this parameter should be a list of User objects.

        projects (List[Project]): it can be an empty list, so one client can be created
            without any project in it. But this parameter should be a list of Project
            objects.
    """

    __auto_name__ = False
    __tablename__ = "Clients"
    __mapper_args__ = {"polymorphic_identity": "Client"}
    client_id: Mapped[int] = mapped_column(
        "id", ForeignKey("Entities.id"), primary_key=True
    )

    users = association_proxy("user_role", "user", creator=lambda n: ClientUser(user=n))

    user_role: Mapped[Optional[List["ClientUser"]]] = relationship(
        back_populates="client",
        cascade="all, delete-orphan",
        primaryjoin="Clients.c.id==Client_Users.c.cid",
        doc="""List of users representing the members of this client.""",
    )

    projects = association_proxy(
        "project_role", "project", creator=lambda p: create_project_client(p)
    )

    project_role: Mapped[Optional[List["ProjectClient"]]] = relationship(
        back_populates="client",
        cascade="all, delete-orphan",
        primaryjoin="Clients.c.id==Project_Clients.c.client_id",
    )

    goods: Mapped[Optional[List["Good"]]] = relationship(
        "Good",
        back_populates="client",
        cascade="all",  # do not include "delete-orphan" we want to keep goods
        # if they are detached on purpose
        primaryjoin="Clients.c.id==Goods.c.client_id",
    )

    def __init__(
        self,
        users: Optional[List["User"]] = None,
        projects: Optional[List["Project"]] = None,
        **kwargs: Optional[Dict[str, Any]],
    ) -> None:
        super(Client, self).__init__(**kwargs)

        if users is None:
            users = []

        if projects is None:
            projects = []

        self.users = users
        self.projects = projects

    def __eq__(self, other: Any) -> bool:
        """Check if the other Client is equal to this once.

        Args:
            other (Any): The other Client instance.

        Returns:
            bool: Returns True, if other object is a Client instance and equal to this
                one.
        """
        return super(Client, self).__eq__(other) and isinstance(other, Client)

    def __hash__(self) -> int:
        """Return the hash value of this instance.

        Because the __eq__ is overridden the __hash__ also needs to be overridden.

        Returns:
            int: The hash value.
        """
        return super(Client, self).__hash__()

    @property
    def to_tjp(self) -> str:
        """Return a TaskJuggler compatible str representation of this Client instance.

        Returns:
            str: The TaskJuggler compatible representation of this Client instance.
        """
        return ""

    @validates("goods")
    def _validate_good(self, key: str, good: "Good") -> "Good":
        """Validate the given good value.

        Args:
            key (str): The name of the validated column.
            good (Good): The good value to be validated.

        Raises:
            TypeError: If the given good is not a :class:`stalker.models.budget.Good`
                instance.

        Returns:
            Good: The validated good value.
        """
        from stalker.models.budget import Good

        if not isinstance(good, Good):
            raise TypeError(
                f"{self.__class__.__name__}.goods should only "
                "contain instances of stalker.models.budget.Good, "
                f"not {good.__class__.__name__}: '{good}'"
            )

        return good


class ClientUser(Base):
    """The association object used in Client-to-User relation.

    Args:
        client (Client): The client which the user is affiliated with.
        user (User): A :class:`.User` instance.
    """

    __tablename__ = "Client_Users"
    user_id: Mapped[int] = mapped_column(
        "uid", ForeignKey("Users.id"), primary_key=True
    )
    user: Mapped["User"] = relationship(
        back_populates="company_role",
        primaryjoin="ClientUser.user_id==User.user_id",
    )
    client_id: Mapped[int] = mapped_column(
        "cid", ForeignKey("Clients.id"), primary_key=True
    )
    client: Mapped["Client"] = relationship(
        back_populates="user_role",
        primaryjoin="ClientUser.client_id==Client.client_id",
    )
    role_id: Mapped[Optional[int]] = mapped_column("rid", ForeignKey("Roles.id"))
    role: Mapped[Optional["Role"]] = relationship(
        primaryjoin="ClientUser.role_id==Role.role_id"
    )

    def __init__(self, client=None, user=None, role=None):
        self.user = user
        self.client = client
        self.role = role

    @validates("client")
    def _validate_client(self, key: str, client: "Client") -> "Client":
        """Validate the given client value.

        Args:
            key (str): The name of the validated column.
            client (Client): The client instance to be validated.

        Raises:
            TypeError: If the given client value is not a :class:`.Client` instance.

        Returns:
            Client: The validated client instance.
        """
        if client is not None:
            if not isinstance(client, Client):
                raise TypeError(
                    f"{self.__class__.__name__}.client should be instance of "
                    "stalker.models.client.Client, "
                    f"not {client.__class__.__name__}: '{client}'"
                )
        return client

    @validates("user")
    def _validate_user(self, key: str, user: "User") -> "User":
        """Validate the given user value.

        Args:
            key (str): The name of the validated column.
            user (User): The user instance to validate.

        Raises:
            TypeError: If the given user value is not a
                :class:`stalker.models.auth.User` instance.

        Returns:
            User: The validated user value.
        """
        if user is not None:
            from stalker.models.auth import User

            if not isinstance(user, User):
                raise TypeError(
                    f"{self.__class__.__name__}.user should be an instance of "
                    "stalker.models.auth.User, "
                    f"not {user.__class__.__name__}: '{user}'"
                )
        return user

    @validates("role")
    def _validate_role(self, key: str, role: "Role") -> "Role":
        """Validate the given role instance.

        Args:
            key (str): The name of the validated column.
            role (Role): The role value to be validated.

        Raises:
            TypeError: If the given role value is not a
                :class:`stalker.models.auth.Role` instance.

        Returns:
            Role: The validated :class:`stalker.models.auth.Role` instance.
        """
        if role is not None:
            from stalker.models.auth import Role

            if not isinstance(role, Role):
                raise TypeError(
                    f"{self.__class__.__name__}.role should be a "
                    "stalker.models.auth.Role instance, "
                    f"not {role.__class__.__name__}: '{role}'"
                )
        return role



# -*- coding: utf-8 -*-
"""Variant related functions and classes are situated here."""

from sqlalchemy import ForeignKey
from sqlalchemy.orm import Mapped, mapped_column

from stalker.models.task import Task


class Variant(Task):
    """A Task derivative to keep track of Variants in a Task hierarchy.

    The basic reason to have the Variant class is to upgrade the variants,
    into a Task derivative so that it is possible to create dependencies
    between different variants and being able to review them individually.

    You see, in previous versions of Stalker, the variants were handled as a
    part of the Version instances with a str attribute. The down side of that
    design was not being able to distinguish any reviews per variant.

    So, when a Model task is approved, all its variant approved all together,
    even if one of the variants were still getting worked on.

    The new design prevents that and gives the variant the level of attention
    they deserved.

    Variants doesn't introduce any new arguments or attributes. They are just
    initialized like any other Tasks.
    """

    __tablename__ = "Variants"
    __mapper_args__ = {"polymorphic_identity": "Variant"}
    variant_id: Mapped[int] = mapped_column(
        "id",
        ForeignKey("Tasks.id"),
        primary_key=True,
    )



# -*- coding: utf-8 -*-
"""Asset related classes."""

import logging
from typing import Any

from sqlalchemy import ForeignKey
from sqlalchemy.orm import Mapped, mapped_column

from stalker import log
from stalker.models.mixins import CodeMixin, ReferenceMixin
from stalker.models.task import Task

logger: logging.Logger = log.get_logger(__name__)
log.set_level(log.logging_level)


class Asset(Task, CodeMixin):
    """The Asset class is the whole idea behind Stalker.

    *Assets* are containers of :class:`.Task` s. And :class:`.Task` s are the
    smallest meaningful part that should be accomplished to complete the
    :class:`.Project`.

    An example could be given as follows; you can create an asset for one of
    the characters in your project. Than you can divide this character asset in
    to :class:`.Task` s. These :class:`.Task` s can be defined by the type of
    the :class:`.Asset`, which is a :class:`.Type` object created specifically
    for :class:`.Asset` (ie. has its :attr:`.Type.target_entity_type` set to
    "Asset"),

    An :class:`.Asset` instance should be initialized with a :class:`.Project`
    instance (as the other classes which are mixed with the
    :class:`.TaskMixin`). And when a :class:`.Project` instance is given then
    the asset will append itself to the :attr:`.Project.assets` list.

    ..versionadded: 0.2.0:
        No more Asset to Shot connection:

        Assets now are not directly related to Shots. Instead a
        :class:`.Version` will reference the Asset and then it is easy to track
        which shots are referencing this Asset by querying with a join of Shot
        Versions referencing this Asset.
    """

    __auto_name__ = False
    __strictly_typed__ = True
    __tablename__ = "Assets"
    __mapper_args__ = {"polymorphic_identity": "Asset"}

    asset_id: Mapped[int] = mapped_column(
        "id", ForeignKey("Tasks.id"), primary_key=True
    )

    def __init__(self, code, **kwargs) -> None:
        kwargs["code"] = code

        super(Asset, self).__init__(**kwargs)
        CodeMixin.__init__(self, **kwargs)
        ReferenceMixin.__init__(self, **kwargs)

    def __eq__(self, other: Any) -> bool:
        """Check the equality.

        Args:
            other (Any): The other object.

        Returns:
            bool: True if the other object equals to this asset.
        """
        return (
            super(Asset, self).__eq__(other)
            and isinstance(other, Asset)
            and self.type == other.type
        )

    def __hash__(self) -> int:
        """Return the hash value of this instance.

        Because the __eq__ is overridden the __hash__ also needs to be overridden.

        Returns:
            int: The hash value.
        """
        return super(Asset, self).__hash__()



# -*- coding: utf-8 -*-
"""Tag related functions and classes are situated here."""
from typing import Any, Dict

from sqlalchemy import ForeignKey
from sqlalchemy.orm import Mapped, mapped_column

from stalker.log import get_logger
from stalker.models.entity import SimpleEntity

logger = get_logger(__name__)


class Tag(SimpleEntity):
    """Use it to create tags for any object available in SOM.

    Doesn't have any other attribute than what is inherited from
    :class:`.SimpleEntity`
    """

    __auto_name__ = False
    __tablename__ = "Tags"
    __mapper_args__ = {"polymorphic_identity": "Tag"}
    tag_id: Mapped[int] = mapped_column(
        "id", ForeignKey("SimpleEntities.id"), primary_key=True
    )

    def __init__(self, **kwargs: Dict[str, Any]) -> None:
        super(Tag, self).__init__(**kwargs)

    def __eq__(self, other: Any) -> bool:
        """Check the equality.

        Args:
            other (Any): The other object.

        Returns:
            bool: True if the other object is a Tag instance and has the same
                attributes.
        """
        return super(Tag, self).__eq__(other) and isinstance(other, Tag)

    def __hash__(self) -> int:
        """Return the hash value for this Tag instance.

        Returns:
            int: The hash of this Tag.
        """
        return super(Tag, self).__hash__()



# -*- coding: utf-8 -*-
"""Shot related functions and classes are situated here."""

from typing import Any, Dict, Optional, TYPE_CHECKING, Union

from sqlalchemy import Float, ForeignKey
from sqlalchemy.exc import OperationalError, UnboundExecutionError
from sqlalchemy.orm import (
    Mapped,
    mapped_column,
    reconstructor,
    relationship,
    synonym,
    validates,
)

from stalker.db.session import DBSession
from stalker.log import get_logger
from stalker.models.format import ImageFormat
from stalker.models.mixins import CodeMixin, ReferenceMixin, StatusMixin
from stalker.models.task import Task

if TYPE_CHECKING:  # pragma: no cover
    from stalker.models.project import Project
    from stalker.models.scene import Scene
    from stalker.models.sequence import Sequence

logger = get_logger(__name__)


class Shot(Task, CodeMixin):
    """Manages Shot related data.

    A shot is a continuous, unbroken sequence of images that makes up a single
    part of a film. Shots are organized into :class:`.Sequence` s and
    :class:`.Scene` s :class:`.Sequence` s group shots together based on time,
    such as montage or flashback. :class:`.Scene` s group shots together based
    on location and narrative, marking where and when a specific story event
    occurs.

    .. warning::

       .. deprecated:: 0.1.2

       Because most of the shots in different projects may going to have
       the same name, which is a kind of a code like SH001, SH012A etc., and
       in Stalker you cannot have two entities with the same name if their
       types are also matching, to guarantee all the shots are going to have
       different names the :attr:`.name` attribute of the Shot instances are
       automatically set to a randomly generated **uuid4** sequence.

    .. note::

       .. versionadded:: 0.1.2

       The name of the shot can be freely set without worrying about clashing
       names.

    .. note::

       .. versionadded:: 0.2.0

       Shot instances now can have their own image format. So you can set up
       different resolutions per shot.

    .. note::

       .. versionadded:: 0.2.0

       Shot instances can now be created with a Project instance only, without
       needing a Sequence instance. Sequences are now a kind of a grouping
       attribute for the Shots. And Shots can have more than one Sequence.

    .. note::

       .. versionadded:: 1.0.0

       Shot instances can only be connected to a single Sequence instance via
       the `Shot.sequence` attribute. Previously, Shots could have multiple
       Sequences, the initial purpose of that was to allow the very very rare
       case of having the same shot appear in two different sequences which
       proved itself being very useless and making things unnecessarily
       complicated. So, it has been removed and Shots can only be connected to
       a single Sequence (Shot <-> Scene relation will follow this in later
       versions/commits).

    .. note::

       .. versionadded:: 1.0.0

       Shot and Scene relation is now many-to-one, meaning a Shot can only be
       connected to a single Scene instance through the `Shot.scene` attribute.

    Two shots with the same :attr:`.code` cannot be assigned to the same
    :class:`.Sequence`.

    .. note::

       .. versionadded:: 0.2.10

       Simplified the implementation of :attr:`.cut_in`, :attr:`.cut_out` and
       :attr:`.cut_duration` attributes. The :attr:`.cut_duration` is always
       the difference between :attr:`.cut_in` and :attr:`.cut_out` and its
       value is only be calculated when it is requested. This greatly
       simplifies the implementation of :attr:`.cut_in` and :attr:`.cut_out`
       attributes.

    The :attr:`.cut_out` and :attr:`.cut_duration` attributes effects each
    other. Setting the :attr:`.cut_out` will change the :attr:`.cut_duration`
    and setting the :attr:`.cut_duration` will change the :attr:`.cut_out`
    value. The default value of the :attr:`.cut_duration` attribute is
    calculated from the :attr:`.cut_in` and :attr:`.cut_out` attributes. If
    both :attr:`.cut_out` and :attr:`.cut_duration` arguments are set to None,
    the :attr:`.cut_duration` defaults to 1 and :attr:`.cut_out` will be set to
    :attr:`.cut_in` + :attr:`.cut_duration`. So the priority of the attributes
    are as follows:

      :attr:`.cut_in` >
      :attr:`.cut_out` >
      :attr:`.cut_duration`

    .. note::

       .. versionadded:: 0.2.4

       :attr:`.handles_at_start` and :attr:`.handles_at_end` attributes.

    .. note::

       .. versionadded:: 0.2.17.2

       Per shot FPS values. It is now possible to change the shot fps by
       setting its :attr:`.fps` attribute. The default values is same with the
       :class:`.Project` .

    Args:
        project (Project): This is the :class:`.Project` instance that this
            shot belongs to. A Shot cannot be created without a Project
            instance.

        sequence (Sequence): This is a :class:`.Sequence` that this shot is
            assigned to. A Shot can be created without having a Sequence
            instance.

        cut_in (int): The in frame number that this shot starts. The default
            value is 1. When the ``cut_in`` is bigger then ``cut_out``, the
            :attr:`.cut_out` attribute is set to :attr:`.cut_in` + 1.

        cut_duration (int): The duration of this shot in frames. It should be
            zero or a positive integer value (natural number?) or . The default
            value is None.

        cut_out (int): The out frame number that this shot ends. If it is given
            as a value lower then the ``cut_in`` parameter, then the
            :attr:`.cut_out` will be recalculated from the existent
            :attr:`.cut_in` :attr:`.cut_duration` attributes. Can be skipped.
            The default value is None.

        image_format (ImageFormat): The image format of this shot. This is an
            optional variable to differentiate the image format per shot. The
            default value is the same with the Project that this Shot belongs
            to.

        fps (float): The FPS of this shot. Default value is the same with the
            :class:`.Project` .
    """

    __auto_name__ = True
    __tablename__ = "Shots"
    __mapper_args__ = {"polymorphic_identity": "Shot"}

    shot_id: Mapped[int] = mapped_column(
        "id",
        ForeignKey("Tasks.id"),
        primary_key=True,
    )

    sequence_id: Mapped[Optional[int]] = mapped_column(ForeignKey("Sequences.id"))

    sequence: Mapped[Optional["Sequence"]] = relationship(
        primaryjoin="Shots.c.sequence_id==Sequences.c.id",
        back_populates="shots",
    )

    scene_id: Mapped[Optional[int]] = mapped_column(ForeignKey("Scenes.id"))

    scene: Mapped[Optional["Scene"]] = relationship(
        primaryjoin="Shots.c.scene_id==Scenes.c.id",
        back_populates="shots",
    )

    image_format_id: Mapped[Optional[int]] = mapped_column(
        ForeignKey("ImageFormats.id")
    )
    _image_format: Mapped[Optional[ImageFormat]] = relationship(
        "ImageFormat",
        primaryjoin="Shots.c.image_format_id==ImageFormats.c.id",
        doc="""The :class:`.ImageFormat` of this shot.

        This value defines the output image format of this shot, should be an
        instance of :class:`.ImageFormat`.
        """,
    )

    # the cut_duration attribute is not going to be stored in the database,
    # only the cut_in and cut_out will be enough to calculate the cut_duration
    cut_in: Mapped[Optional[int]] = mapped_column(
        doc="The start frame of this shot. It is the start frame of the "
        "playback range in the application (Maya, Nuke etc.).",
        default=1,
    )
    cut_out: Mapped[Optional[int]] = mapped_column(
        doc="The end frame of this shot. It is the end frame of the "
        "playback range in the application (Maya, Nuke etc.).",
        default=1,
    )

    source_in: Mapped[Optional[int]] = mapped_column(
        doc="The start frame of the used range, should be in between"
        ":attr:`.cut_in` and :attr:`.cut_out`",
    )
    source_out: Mapped[Optional[int]] = mapped_column(
        doc="The end frame of the used range, should be in between"
        ":attr:`.cut_in and :attr:`.cut_out`",
    )
    record_in: Mapped[Optional[int]] = mapped_column(
        doc="The start frame in the Editors timeline specifying the start "
        "frame general placement of this shot.",
    )

    _fps: Mapped[Optional[float]] = mapped_column(
        "fps",
        Float(precision=3),
        doc="""The fps of the project.

        It is a float value, any other types will be converted to float. The
        default value is equal to :attr:`stalker.models.project..Project.fps`.
        """,
    )

    def __init__(
        self,
        code: Optional[str] = None,
        project: Optional["Project"] = None,
        sequence: Optional["Sequence"] = None,
        scene: Optional["Scene"] = None,
        cut_in: Optional[int] = None,
        cut_out: Optional[int] = None,
        source_in: Optional[int] = None,
        source_out: Optional[int] = None,
        record_in: Optional[int] = None,
        image_format: Optional[ImageFormat] = None,
        fps: Optional[float] = None,
        **kwargs: Dict[str, Any],
    ) -> None:
        kwargs["project"] = project
        kwargs["code"] = code

        self._updating_cut_in_cut_out = False

        super(Shot, self).__init__(**kwargs)
        ReferenceMixin.__init__(self, **kwargs)
        StatusMixin.__init__(self, **kwargs)
        CodeMixin.__init__(self, **kwargs)

        self.sequence = sequence
        self.scene = scene
        self.image_format = image_format

        if cut_in is None:
            if cut_out is not None:
                cut_in = cut_out

        if cut_out is None:
            if cut_in is not None:
                cut_out = cut_in

        # if both are None set them to default values
        if cut_in is None and cut_out is None:
            cut_in = 1
            cut_out = 1

        self.cut_in = cut_in
        self.cut_out = cut_out

        if source_in is None:
            source_in = self.cut_in

        if source_out is None:
            source_out = self.cut_out

        self.source_in = source_in
        self.source_out = source_out
        self.record_in = record_in

        self.fps = fps

    @reconstructor
    def __init_on_load__(self) -> None:
        """Initialize on DB load."""
        super(Shot, self).__init_on_load__()
        self._updating_cut_in_cut_out = False

    def __repr__(self) -> str:
        """Return the string representation of this Shot instance.

        Returns:
            str: The string representation of this Shot instance.
        """
        return f"<{self.entity_type} ({self.name}, {self.code})>"

    def __eq__(self, other: Any) -> bool:
        """Check the equality.

        Args:
            other (Any): The other object.

        Returns:
            bool: True if the other object is a Shot instance and has the same code and
                project.
        """
        return (
            isinstance(other, Shot)
            and self.code == other.code
            and self.project == other.project
        )

    def __hash__(self) -> int:
        """Return the hash value of this instance.

        Because the __eq__ is overridden the __hash__ also needs to be overridden.

        Returns:
            int: The hash value.
        """
        return super(Shot, self).__hash__()

    @classmethod
    def _check_code_availability(cls, code: str, project: "Project") -> bool:
        """Check if the given code is available in the given project.

        Args:
            code (str): The code to check the availability of.
            project (Project): The stalker.models.project.Project instance that this
                shot is a part of.

        Raises:
            TypeError: If the code is not a str.

        Returns:
            bool: True if the given code is available, False otherwise.
        """
        if not project or not code:
            return True

        if not isinstance(code, str):
            raise TypeError(
                "code should be a string containing a shot code, "
                f"not {code.__class__.__name__}: '{code}'"
            )

        from stalker import Project

        if not isinstance(project, Project):
            raise TypeError(
                "project should be a Project instance, "
                f"not {project.__class__.__name__}: '{project}'"
            )

        try:
            logger.debug("Try checking Shot.code with SQL expression.")
            with DBSession.no_autoflush:
                return (
                    Shot.query.filter(Shot.project == project)
                    .filter(Shot.code == code)
                    .first()
                    is None
                )
        except (UnboundExecutionError, OperationalError):
            # Fallback to Python
            logger.debug("SQL expression failed, falling back to Python!")
            for t in project.tasks:
                if isinstance(t, Shot) and t.code == code:
                    return False
        return True

    def _fps_getter(self) -> float:
        """Return the fps value either from the Project or from the _fps attribute.

        Returns:
            float: The fps attribute value.
        """
        if self._fps is None:
            return self.project.fps
        else:
            return self._fps

    def _fps_setter(self, fps: float) -> None:
        """Set the fps value.

        Args:
            fps (float): The fps value to set the fps attribute to.
        """
        self._fps = self._validate_fps(fps)

    fps: Mapped[Optional[float]] = synonym(
        "_fps",
        descriptor=property(_fps_getter, _fps_setter),
        doc="The fps of this shot.",
    )

    def _validate_fps(self, fps: Union[int, float]) -> float:
        """Validate the given fps value.

        Args:
            fps (Union[int, float]): Either an integer or float value to used as the
                fps.

        Raises:
            TypeError: If the given `fps` value is not an integer or float.
            ValueError: If the `fps` value is smaller or equal to 0.

        Returns:
            float: The validated fps value.
        """
        if fps is None:
            # fps = self.project.fps
            return None

        if not isinstance(fps, (int, float)):
            raise TypeError(
                "{}.fps should be a positive float or int, not {}: '{}'".format(
                    self.__class__.__name__, fps.__class__.__name__, fps
                )
            )

        fps = float(fps)
        if fps <= 0:
            raise ValueError(
                "{}.fps should be a positive float or int, not {}".format(
                    self.__class__.__name__, fps
                )
            )
        return float(fps)

    @validates("cut_in")
    def _validate_cut_in(self, key: str, cut_in: int) -> int:
        """Validate the cut_in value.

        Args:
            key (str): The name of the validated column.
            cut_in (int): The `cut_in` value to be validated.

        Raises:
            TypeError: If the given `cut_in` value is not an integer.

        Returns:
            int: The validated `cut_in` value.
        """
        if not isinstance(cut_in, int):
            raise TypeError(
                f"{self.__class__.__name__}.cut_in should be an int, "
                f"not {cut_in.__class__.__name__}: '{cut_in}'"
            )

        if self.cut_out is not None and not self._updating_cut_in_cut_out:
            if cut_in > self.cut_out:
                # lock the attribute update
                self._updating_cut_in_cut_out = True
                self.cut_out = cut_in
                self._updating_cut_in_cut_out = False

        return cut_in

    @validates("cut_out")
    def _validate_cut_out(self, key: str, cut_out: int) -> int:
        """Validate the cut_out value.

        Args:
            key (str): The name of the validated column.
            cut_out (int): The `cut_out` value to be validated.

        Raises:
            TypeError: If the `cut_out` value is not an integer.

        Returns:
            int: The validated `cut_out` value.
        """
        if not isinstance(cut_out, int):
            raise TypeError(
                f"{self.__class__.__name__}.cut_out should be an int, "
                f"not {cut_out.__class__.__name__}: '{cut_out}'"
            )

        if (
            self.cut_in is not None
            and not self._updating_cut_in_cut_out
            and cut_out < self.cut_in
        ):
            # lock the attribute update
            self._updating_cut_in_cut_out = True
            self.cut_in = cut_out
            self._updating_cut_in_cut_out = False

        return cut_out

    @validates("source_in")
    def _validate_source_in(self, key: str, source_in: int) -> int:
        """Validate the source_in value.

        Args:
            key (str): The name of the validated column.
            source_in (int): The `source_in` value to be validated.

        Raises:
            TypeError: If the `source_in` value is not an int.
            ValueError: If the given `source_in` value is smaller than the `cut_in`
                attribute value.
            ValueError: If the given `source_in` value is larger than the `cut_out`
                attribute value.
            ValueError: If a `source_out` is given before and the `source_in` value is
                larger than the `source_out` attribute value.

        Returns:
            int: The validated `source_in` value.
        """
        if not isinstance(source_in, int):
            raise TypeError(
                f"{self.__class__.__name__}.source_in should be an int, "
                f"not {source_in.__class__.__name__}: '{source_in}'"
            )

        if source_in < self.cut_in:
            raise ValueError(
                "{cls}.source_in cannot be smaller than "
                "{cls}.cut_in, cut_in: {cut_in} where as "
                "source_in: {source_in}".format(
                    cls=self.__class__.__name__,
                    cut_in=self.cut_in,
                    source_in=source_in,
                )
            )

        if source_in > self.cut_out:
            raise ValueError(
                "{cls}.source_in cannot be bigger than "
                "{cls}.cut_out, cut_out: {cut_out} where as "
                "source_in: {source_in}".format(
                    cls=self.__class__.__name__,
                    cut_out=self.cut_out,
                    source_in=source_in,
                )
            )

        if self.source_out and source_in > self.source_out:
            raise ValueError(
                "{cls}.source_in cannot be bigger than "
                "{cls}.source_out, source_in: {source_in} where "
                "as source_out: {source_out}".format(
                    cls=self.__class__.__name__,
                    source_out=self.source_out,
                    source_in=source_in,
                )
            )

        return source_in

    @validates("source_out")
    def _validate_source_out(self, key: str, source_out: int) -> int:
        """Validate the source_out value.

        Args:
            key (str): The name of the validated column.
            source_out (int): The source_out value to be validated.

        Raises:
            TypeError: If the source_out is not an integer.
            ValueError: If the source_out is smaller than the cut_in attribute value.
            ValueError: If the source_out is larger than the cut_out attribute value.
            ValueError: If the source_in is not None and source_out is smaller than the
                source_in value.

        Returns:
            int: The validated source_out value.
        """
        if not isinstance(source_out, int):
            raise TypeError(
                f"{self.__class__.__name__}.source_out should be an int, "
                f"not {source_out.__class__.__name__}: '{source_out}'"
            )

        if source_out < self.cut_in:
            raise ValueError(
                "{cls}.source_out cannot be smaller than "
                "{cls}.cut_in, cut_in: {cut_in} where as "
                "source_out: {source_out}".format(
                    cls=self.__class__.__name__,
                    cut_in=self.cut_in,
                    source_out=source_out,
                )
            )

        if source_out > self.cut_out:
            raise ValueError(
                "{cls}.source_out cannot be bigger than "
                "{cls}.cut_out, cut_out: {cut_out} where as "
                "source_out: {source_out}".format(
                    cls=self.__class__.__name__,
                    cut_out=self.cut_out,
                    source_out=source_out,
                )
            )

        if self.source_in and source_out < self.source_in:
            raise ValueError(
                "{cls}.source_out cannot be smaller than "
                "{cls}.source_in, source_in: {source_in} where "
                "as source_out: {source_out}".format(
                    cls=self.__class__.__name__,
                    source_in=self.source_in,
                    source_out=source_out,
                )
            )

        return source_out

    # @validates('record_in')
    # def _validate_record_in(self, key, record_in):
    #     """validates the given record_in value
    #     """
    #     # we don't really care about the record in value right now.
    #     # it can be set to anything
    #     return record_in

    @property
    def cut_duration(self) -> int:
        """Return the cut_duration property value.

        Returns:
            int: The cut_duration property value.
        """
        return self.cut_out - self.cut_in + 1

    @cut_duration.setter
    def cut_duration(self, cut_duration: int) -> None:
        """Set the cut_duration attribute.

        Args:
            cut_duration (int): The cut_duration value to be validated.

        Raises:
            TypeError: If the given cut_duration value is not an integer.
            ValueError: If the given cut_duration value is not a positive integer.
        """
        if not isinstance(cut_duration, int):
            raise TypeError(
                f"{self.__class__.__name__}.cut_duration should be a positive "
                "integer value, "
                f"not {cut_duration.__class__.__name__}: '{cut_duration}'"
            )

        if cut_duration < 1:
            raise ValueError(
                f"{self.__class__.__name__}.cut_duration cannot be set to "
                "zero or a negative value"
            )

        # always extend or contract the shot from end
        self.cut_out = self.cut_in + cut_duration - 1

    @validates("sequence")
    def _validate_sequence(self, key: str, sequence: "Sequence") -> "Sequence":
        """Validate the given sequence value.

        Args:
            key (str): The name of the validated column.
            sequence (Sequence): The sequence value to validate.

        Raises:
            TypeError: If the given sequence value is not a Sequence instance.

        Returns:
            Sequence: The validated Sequence instance.
        """
        from stalker.models.sequence import Sequence

        if sequence is not None and not isinstance(sequence, Sequence):
            raise TypeError(
                f"{self.__class__.__name__}.sequence should be a "
                "stalker.models.sequence.Sequence instance, "
                f"not {sequence.__class__.__name__}: '{sequence}'"
            )
        return sequence

    @validates("scene")
    def _validate_scene(self, key: str, scene: "Scene") -> "Scene":
        """Validate the given scene value.

        Args:
            key (str): The name of the validated column.
            scene (Scene): The scene value to validate.

        Raises:
            TypeError: If the given scene is not a Scene instance.

        Returns:
            Scene: The validated Scene instance.
        """
        from stalker.models.scene import Scene

        if scene is not None and not isinstance(scene, Scene):
            raise TypeError(
                f"{self.__class__.__name__}.scene should be a "
                "stalker.models.scene.Scene instance, "
                f"not {scene.__class__.__name__}: '{scene}'"
            )
        return scene

    def _image_format_getter(self) -> ImageFormat:
        """Return image_format value from the Project or from the _image_format attr.

        Returns:
            ImageFormat: The ImageFormat instance from image_format attribute or from
                the related Project's image_format attribute.
        """
        if self._image_format is None:
            return self.project.image_format
        else:
            return self._image_format

    def _image_format_setter(self, imf: ImageFormat) -> None:
        """Set the image_format value.

        Args:
            imf (ImageFormat): The ImageFormat instance to set the image_format
                attribute value.
        """
        self._image_format = self._validate_image_format(imf)

    image_format: Mapped[Optional[ImageFormat]] = synonym(
        "_image_format",
        descriptor=property(_image_format_getter, _image_format_setter),
        doc="The image_format of this shot. Set it to None to re-sync with "
        "Project.image_format.",
    )

    def _validate_image_format(
        self, imf: Union[None, ImageFormat]
    ) -> Union[None, ImageFormat]:
        """Validate the given imf value.

        Args:
            imf (ImageFormat): The ImageFormat instance to validate.

        Raises:
            TypeError: If the given imf value is not an ImageFormat instance.

        Returns:
            ImageFormat: The validated ImageFormat instance.
        """
        if imf is None:
            # do not set it to anything it will automatically use the project
            # image format
            return None

        if not isinstance(imf, ImageFormat):
            raise TypeError(
                f"{self.__class__.__name__}.image_format should be an instance of "
                "stalker.models.format.ImageFormat, "
                f"not {imf.__class__.__name__}: '{imf}'"
            )

        return imf

    @validates("code")
    def _validate_code(self, key: str, code: str) -> str:
        """Validate the given code value.

        Args:
            key (str): The name of the validated column.
            code (str): The code to validate.

        Raises:
            ValueError: If the code is not available.

        Returns:
            str: The validated code value.
        """
        code = super(Shot, self)._validate_code(key, code)

        # check code uniqueness
        if code != self.code and not self._check_code_availability(code, self.project):
            raise ValueError(f"There is a Shot with the same code: {code}")

        return code



# -*- coding: utf-8 -*-
"""Tests for the exceptions module."""
import pytest

from stalker.exceptions import (
    CircularDependencyError,
    DependencyViolationError,
    LoginError,
    OverBookedError,
    StatusError,
)


def test_login_error_is_working_as_expected():
    """LoginError is working as expected."""
    test_message = "testing LoginError"
    with pytest.raises(LoginError) as cm:
        raise LoginError(test_message)

    assert str(cm.value) == test_message


def test_circular_dependency_error_is_working_as_expected():
    """CircularDependencyError is working as expected."""
    test_message = "testing CircularDependencyError"
    with pytest.raises(CircularDependencyError) as cm:
        raise CircularDependencyError(test_message)

    assert str(cm.value) == test_message


def test_over_booked_error_is_working_as_expected():
    """OverBookedError is working as expected."""
    test_message = "testing OverBookedError"
    with pytest.raises(OverBookedError) as cm:
        raise OverBookedError(test_message)

    assert str(cm.value) == test_message


def test_status_error_is_working_as_expected():
    """StatusError is working as expected."""
    test_message = "testing StatusError"
    with pytest.raises(StatusError) as cm:
        raise StatusError(test_message)

    assert str(cm.value) == test_message


def test_dependency_violation_error_is_working_as_expected():
    """DependencyViolationError is working as expected."""
    test_message = "testing DependencyViolationError"
    with pytest.raises(DependencyViolationError) as cm:
        raise DependencyViolationError(test_message)

    assert str(cm.value) == test_message



# -*- coding: utf-8 -*-
"""Sequence related function and classes are situated here."""

from typing import Any, Dict, List, Optional, TYPE_CHECKING

from sqlalchemy import ForeignKey
from sqlalchemy.orm import Mapped, mapped_column, relationship, validates

from stalker.log import get_logger
from stalker.models.mixins import CodeMixin, ReferenceMixin
from stalker.models.task import Task

if TYPE_CHECKING:  # pragma: no cover
    from stalker.models.shot import Shot

logger = get_logger(__name__)


class Sequence(Task, CodeMixin):
    """Stores data about Sequences.

    Sequences are a way of grouping the Shots according to their temporal
    position to each other.

    **Initialization**

    .. warning::

       .. deprecated:: 0.2.0

       Sequences do not have a lead anymore. Use the :class:`.Task.responsible`
       attribute of the super (:class:`.Task`).
    """

    __auto_name__ = False
    __tablename__ = "Sequences"
    __mapper_args__ = {"polymorphic_identity": "Sequence"}
    sequence_id: Mapped[int] = mapped_column(
        "id",
        ForeignKey("Tasks.id"),
        primary_key=True,
    )

    shots: Mapped[Optional[List["Shot"]]] = relationship(
        primaryjoin="Shots.c.sequence_id==Sequences.c.id",
        back_populates="sequence",
        doc="""The :class:`.Shot` s assigned to this Sequence.

        It is a list of :class:`.Shot` instances.
        """,
    )

    def __init__(self, **kwargs: Dict[str, Any]) -> None:
        super(Sequence, self).__init__(**kwargs)

        # call the mixin __init__ methods
        ReferenceMixin.__init__(self, **kwargs)
        CodeMixin.__init__(self, **kwargs)
        self.shots = []

    @validates("shots")
    def _validate_shots(self, key: str, shot: "Shot") -> "Shot":
        """Validate the given shot value.

        Args:
            key (str): The name of the validated column.
            shot (Shot): The Shot instance to validate.

        Raises:
            TypeError: If the given shot is not a Shot instance.

        Returns:
            Shot: The validated shot value.
        """
        from stalker.models.shot import Shot

        if not isinstance(shot, Shot):
            raise TypeError(
                f"{self.__class__.__name__}.shots should only contain "
                "instances of stalker.models.shot.Shot, "
                f"not {shot.__class__.__name__}: '{shot}'"
            )
        return shot

    def __eq__(self, other: Any) -> bool:
        """Check the equality.

        Args:
            other (Any): The other object.

        Returns:
            bool: True if the other object is a Sequence instance and has the same
                attributes.
        """
        return isinstance(other, Sequence) and super(Sequence, self).__eq__(other)

    def __hash__(self) -> int:
        """Return the hash value of this instance.

        Because the __eq__ is overridden the __hash__ also needs to be overridden.

        Returns:
            int: The hash value.
        """
        return super(Sequence, self).__hash__()



# -*- coding: utf-8 -*-
"""Wiki related functions and classes are situated here."""
from typing import Any, Dict, Optional, TYPE_CHECKING, Union

from sqlalchemy import ForeignKey, Text
from sqlalchemy.orm import Mapped, mapped_column, validates

from stalker import Entity, ProjectMixin

if TYPE_CHECKING:  # pragma: no cover
    from stalker.models.project import Project


class Page(Entity, ProjectMixin):
    """A simple Wiki page implementation.

    Wiki in Stalker are managed per Project. That is, all Wiki pages are
    related to a Project.

    Stalker wiki pages are very simple in terms of data it holds. It has only
    one :attr:`.title` and one :attr:`.content` an some usual audit info coming
    from :class:`.SimpleEntity` and a :attr:`.project` coming from
    :class:`.ProjectMixin`.

    Args:
        title (str): The title of this Page.
        content (str): The content of this page. Can contain any kind of string
            literals including HTML tags etc.
    """

    __auto_name__ = True
    __tablename__ = "Pages"
    __mapper_args__ = {"polymorphic_identity": "Page"}
    page_id: Mapped[int] = mapped_column(
        "id",
        ForeignKey("Entities.id"),
        primary_key=True,
    )

    title: Mapped[Optional[str]] = mapped_column(Text)
    content: Mapped[Optional[str]] = mapped_column(Text)

    def __init__(
        self,
        title: str = "",
        content: str = "",
        project: Optional["Project"] = None,
        **kwargs: Dict[str, Any],
    ) -> None:
        kwargs["project"] = project
        super(Page, self).__init__(**kwargs)
        ProjectMixin.__init__(self, **kwargs)

        self.title = title
        self.content = content

    @validates("title")
    def _validate_title(self, key: str, title: str) -> str:
        """Validate the given title value.

        Args:
            key (str): The name of the validated column.
            title (str): The title value to be validated.

        Raises:
            TypeError: If the given title is not a string.
            ValueError: If the title is an empty string.

        Returns:
            str: The validated title value.
        """
        if not isinstance(title, str):
            raise TypeError(
                "{}.title should be a string, not {}: '{}'".format(
                    self.__class__.__name__, title.__class__.__name__, title
                )
            )

        if not title:
            raise ValueError(f"{self.__class__.__name__}.title cannot be empty")

        return title

    @validates("content")
    def _validate_content(self, key: str, content: Union[None, str]) -> str:
        """Validate the given content value.

        Args:
            key (str): The name of the validated column.
            content (Union[None, str]): The content value to be validated.

        Raises:
            TypeError: If the content is not None and not str.

        Returns:
            str: The validated content value.
        """
        content = "" if content is None else content
        if not isinstance(content, str):
            raise TypeError(
                "{}.content should be a string, not {}: '{}'".format(
                    self.__class__.__name__, content.__class__.__name__, content
                )
            )
        return content



# -*- coding: utf-8 -*-
"""Department related classes and functions are situated here."""

from typing import Any, Dict, List, Optional, Union

from sqlalchemy import ForeignKey
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.orm import Mapped, mapped_column, relationship, validates

from stalker.db.declarative import Base
from stalker.log import get_logger
from stalker.models.auth import Role, User
from stalker.models.entity import Entity

logger = get_logger(__name__)


class Department(Entity):
    """The departments that forms the studio itself.

    The information that a Department object holds is like:

      * The members of the department
      * and all the other things those are inherited from the AuditEntity class

    Two Department object considered the same if they have the same name, the
    the users list is not important, a "Modeling" department
    should of course be the same with another department which has the name
    "Modeling" again.

    so creating a department object needs the following parameters:

    Args:
        users (List[User]): it can be an empty list, so one department can be
            created without any member in it. But this parameter should be a list
            of User objects.
    """

    __auto_name__ = False
    __tablename__ = "Departments"
    __mapper_args__ = {"polymorphic_identity": "Department"}
    department_id: Mapped[int] = mapped_column(
        "id", ForeignKey("Entities.id"), primary_key=True
    )

    users = association_proxy(
        "user_role", "user", creator=lambda u: DepartmentUser(user=u)
    )

    user_role: Mapped[Optional[List["DepartmentUser"]]] = relationship(
        back_populates="department",
        cascade="all, delete-orphan",
        primaryjoin="Departments.c.id==Department_Users.c.did",
        doc="""List of users representing the members of this department.""",
    )

    def __init__(
        self, users: Optional[List[User]] = None, **kwargs: Optional[Dict[str, Any]]
    ) -> None:
        super(Department, self).__init__(**kwargs)

        if users is None:
            users = []

        self.users = users

    def __eq__(self, other: Any) -> bool:
        """Check if the other is equal to this one.

        Args:
            other (Any): The other Department instance.

        Returns:
            bool: True if the other object is also a Department and all the attributes
                are equal.
        """
        return super(Department, self).__eq__(other) and isinstance(other, Department)

    def __hash__(self) -> int:
        """Return the hash value of this instance.

        Because the __eq__ is overridden the __hash__ also needs to be overridden.

        Returns:
            int: The hash value.
        """
        return super(Department, self).__hash__()

    @validates("user_role")
    def _validate_user_role(
        self, key: str, user_role: "DepartmentUser"
    ) -> "DepartmentUser":
        """Validate the given user_role value.

        Args:
            key (str): The name of the validated column.
            user_role (DepartmentUser): The user_role value to be validated.

        Returns:
            DepartmentUser: The validated user_role value.
        """
        return user_role

    @property
    def to_tjp(self) -> str:
        """Output a TaskJuggler compatible representation.

        Returns:
            str: The TaskJuggler compatible representation.
        """
        tab = "    "
        indent = tab
        tjp = f'resource {self.tjp_id} "{self.tjp_id}" {{'
        for resource in self.users:
            tjp += "\n"
            tjp += "\n".join(f"{indent}{line}" for line in resource.to_tjp.split("\n"))
        tjp += "\n}"
        return tjp


# DEPARTMENTS_USERS
class DepartmentUser(Base):
    """The association object used in Department-to-User relation."""

    __tablename__ = "Department_Users"

    user_id: Mapped[int] = mapped_column(
        "uid", ForeignKey("Users.id"), primary_key=True
    )
    user: Mapped["User"] = relationship(
        back_populates="department_role",
        primaryjoin="DepartmentUser.user_id==User.user_id",
        uselist=False,
    )
    department_id: Mapped[int] = mapped_column(
        "did", ForeignKey("Departments.id"), primary_key=True
    )
    department: Mapped[Department] = relationship(
        back_populates="user_role",
        primaryjoin="DepartmentUser.department_id==Department.department_id",
        uselist=False,
    )
    role_id: Mapped[Optional[int]] = mapped_column("rid", ForeignKey("Roles.id"))
    role: Mapped[Role] = relationship(
        primaryjoin="DepartmentUser.role_id==Role.role_id"
    )

    def __init__(self, department=None, user=None, role=None):
        self.department = department
        self.user = user
        self.role = role

    @validates("department")
    def _validate_department(
        self, key: str, department: Union[None, Department]
    ) -> Union[None, Department]:
        """Validate the given department value.

        Args:
            key (str): The name of the validated column.
            department (Department): The department value to be validated.

        Raises:
            TypeError: If the given user value is not a :class:`.Department` instance.

        Returns:
            Department: The validated department value.
        """
        if department is not None:
            # check if it is instance of Department object
            if not isinstance(department, Department):
                raise TypeError(
                    f"{self.__class__.__name__}.department should be a "
                    "stalker.models.department.Department instance, "
                    f"not {department.__class__.__name__}: '{department}'"
                )
        return department

    @validates("user")
    def _validate_user(
        self, key: str, user: Union[None, "User"]
    ) -> Union[None, "User"]:
        """Validate the given user value.

        Args:
            key (str): The name of the validated column.
            user (User): The user value to be validated.

        Raises:
            TypeError: If the given user value is not a
                :class:`stalker.models.auth.User` instance.

        Returns:
            User: The validated user value.
        """
        if user is not None:
            if not isinstance(user, User):
                raise TypeError(
                    f"{self.__class__.__name__}.user should be a "
                    "stalker.models.auth.User instance, "
                    f"not {user.__class__.__name__}: '{user}'"
                )
        return user

    @validates("role")
    def _validate_role(self, key: str, role: Union[None, Role]) -> Union[None, Role]:
        """Validate the given role instance.

        Args:
            key (str): The name of the validated column.
            role (Union[None, Role]): The role value to be validated.

        Raises:
            TypeError: If the given role value is not a
                :class:`stalker.models.auth.Role` instance.

        Returns:
            Union[None, Role]: The validated role value.
        """
        if role is not None:
            if not isinstance(role, Role):
                raise TypeError(
                    f"{self.__class__.__name__}.role should be a "
                    "stalker.models.auth.Role instance, "
                    f"not {role.__class__.__name__}: '{role}'"
                )
        return role



# -*- coding: utf-8 -*-
import stalker.db.setup
from stalker import db
from stalker.db.session import DBSession
from stalker import (
    Asset,
    FilenameTemplate,
    ImageFormat,
    Repository,
    Project,
    Shot,
    Structure,
    Task,
    Type,
    User,
    Version,
)
from stalker.models.enum import TimeUnit


def test_readme_tutorial_code(setup_sqlite3):
    """the tutorial code in README.rst."""
    stalker.db.setup.setup()
    stalker.db.setup.init()

    assert str(DBSession.connection().engine.url) == "sqlite://"

    me = User(
        name="Erkan Ozgur Yilmaz",
        login="erkanozgur",
        email="my_email@gmail.com",
        password="secretpass",
    )

    # Save the user to database
    DBSession.save(me)

    repo = Repository(
        name="Commercial Projects Repository",
        code="CPR",
        windows_path="Z:/Projects",
        linux_path="/mnt/Z/Projects",
        macos_path="/Volumes/Z/Projects",
    )

    task_template = FilenameTemplate(
        name="Standard Task Filename Template",
        target_entity_type="Task",  # This is for files saved for Tasks
        path="{{project.repository.path}}/{{project.code}}/"
        "{%- for parent_task in parent_tasks -%}"
        "{{parent_task.nice_name}}/"
        "{%- endfor -%}",  # This is Jinja2 template code
        filename='{{version.nice_name}}_v{{"%03d"|format(version.version_number)}}',
    )

    standard_folder_structure = Structure(
        name="Standard Project Folder Structure",
        templates=[task_template],
        custom_template="{{project.code}}/References",  # If you need extra folders
    )

    new_project = Project(
        name="Test Project",
        code="TP",
        structure=standard_folder_structure,
        repositories=[repo],  # if you have more than one repository you can do it
    )

    hd1080 = ImageFormat(name="1080p", width=1920, height=1080)

    new_project.image_format = hd1080

    # Save the project and all the other data it is connected to it
    DBSession.save(new_project)

    # define Character asset type
    char_type = Type(name="Character", code="CHAR", target_entity_type="Asset")

    character1 = Asset(
        name="Character 1", code="CHAR1", type=char_type, project=new_project
    )

    # Save the Asset
    DBSession.save(character1)

    model = Task(name="Model", parent=character1)

    rigging = Task(
        name="Rig",
        parent=character1,
        depends_on=[model],  # For project management, define that Rig cannot start
        # before Model ends.
    )

    # Save the new tasks
    DBSession.save([model, rigging])

    # A shot and some tasks for it
    shot = Shot(name="SH001", code="SH001", project=new_project)

    # Save the Shot
    DBSession.save(shot)

    animation = Task(
        name="Animation",
        parent=shot,
    )

    lighting = Task(
        name="Lighting",
        parent=shot,
        depends_on=[animation],  # Lighting cannot start before Animation ends,
        schedule_timing=1,
        schedule_unit=TimeUnit.Day,  # The task expected to take 1 day to complete
        resources=[me],
    )
    DBSession.save([animation, lighting])

    new_version = Version(task=animation)
    new_version.generate_path()  # to render the naming convention template
    new_version.extension = ".ma"  # let's say that we have created under Maya
    DBSession.save(new_version)

    path = new_version.generate_path(extension=".ma")
    assert str(path) == f"{repo.path}TP/SH001/Animation/SH001_Animation_v001.ma"
    assert new_version.version_number == 1

    new_version2 = Version(task=animation)
    DBSession.save(new_version2)

    # to render the naming convention template
    # let's say that we have created under Maya
    # path = new_version2.generate_path(extension = ".ma")
    assert new_version2.version_number == 2



# -*- coding: utf-8 -*-
"""Utilities are situated here."""
import calendar
from datetime import datetime, timedelta
from typing import Any, Generator, Union

import pytz

from stalker.exceptions import CircularDependencyError
from stalker.models.enum import TraversalDirection


def make_plural(name: str) -> str:
    """Return the plural version of the given name argument.

    Args:
        name (str): The name to make plural.

    Returns:
        str: The plural version of the given name.
    """
    plural_name = name + "s"

    if name[-1] == "y":
        plural_name = name[:-1] + "ies"
    elif name[-2:] == "ch":
        plural_name = name + "es"
    elif name[-1] == "f":
        plural_name = name[:-1] + "ves"
    elif name[-1] == "s":
        plural_name = name + "es"

    return plural_name


def walk_hierarchy(
    entity: Any,
    attr: str,
    method: Union[int, str, TraversalDirection] = TraversalDirection.DepthFirst,
) -> Generator[Any, Any, Any]:
    """Walk the entity hierarchy over the given attribute and yield the entities found.

    It doesn't check for cycle, so if the attribute is not acyclic then this
    function will not find an exit point.

    The default method is Depth First Search (DFS), to walk with Breadth First
    Search (BFS) set the direction to :attr:`.TraversalDirection.BreadthFirst`.

    Args:
        entity (Any): Starting Entity.
        attr (str): The attribute name to walk over.
        method (Union[int, str, TraversalDirection]): Use TraversalDirection
            enum values, or one of the values listed here ["DepthFirst",
            "BreadthFirst", 0, 1]. The default is
            :attr:`.TraversalDirection.DepthFirst`.

    Yields:
        Any: List any entities found while traversing the hierarchy.
    """
    entity_to_visit = [entity]
    method = TraversalDirection.to_direction(method)
    if method == TraversalDirection.DepthFirst:
        while len(entity_to_visit):
            current_entity = entity_to_visit.pop(0)
            for child in reversed(getattr(current_entity, attr)):
                entity_to_visit.insert(0, child)
            yield current_entity
    else:  # TraversalDirection.BreadthFirst
        while len(entity_to_visit):
            current_entity = entity_to_visit.pop(0)
            entity_to_visit.extend(getattr(current_entity, attr))
            yield current_entity


def check_circular_dependency(entity: Any, other_entity: Any, attr_name: str) -> None:
    """Check circular dependency.

    Check if entity and other_entity are in circular dependency over the attr with the
    name attr_name.

    Args:
        entity (Any): Any Python object.
        other_entity (Any): Any Python object.
        attr_name (str): The name of the attribute to check the circular dependency of.

    Raises:
        CircularDependencyError: If the entities are in circular dependency over the
            attr with the name attr_name.
    """
    for e in walk_hierarchy(entity, attr_name):
        if e is other_entity:
            raise CircularDependencyError(
                "{entity_name} ({entity_class}) and "
                "{other_entity_name} ({other_entity_class}) are in a "
                'circular dependency in their "{attr_name}" attribute'.format(
                    entity_name=entity,
                    entity_class=entity.__class__.__name__,
                    other_entity_name=other_entity,
                    other_entity_class=other_entity.__class__.__name__,
                    attr_name=attr_name,
                )
            )


def utc_to_local(utc_datetime: datetime) -> datetime:
    """Convert utc time to local time.

    Based on the answer of J.F. Sebastian on
    http://stackoverflow.com/questions/4563272/how-to-convert-a-python-utc-datetime-to-a-local-datetime-using-only-python-stand/13287083#13287083

    Args:
        utc_datetime (datetime): The UTC datetime instance.

    Returns:
        datetime: The local datetime instance.
    """
    # get integer timestamp to avoid precision lost
    timestamp = calendar.timegm(utc_datetime.timetuple())
    local_dt = datetime.fromtimestamp(timestamp)
    return local_dt.replace(microsecond=utc_datetime.microsecond)


def local_to_utc(local_datetime: datetime) -> datetime:
    """Convert local datetime to utc datetime.

    Based on the answer of J.F. Sebastian on
    http://stackoverflow.com/questions/4563272/how-to-convert-a-python-utc-datetime-to-a-local-datetime-using-only-python-stand/13287083#13287083

    Args:
        local_datetime (datetime): The local `datetime` instance.

    Returns:
        datetime: The UTC datetime instance.
    """
    # get the utc_datetime as if the local_datetime is utc and calculate the timezone
    # difference and add it to the local datetime object
    return local_datetime - (utc_to_local(local_datetime) - local_datetime)


def datetime_to_millis(dt: datetime) -> int:
    """Calculate the milliseconds since epoch for the given datetime value.

    This is used as the default JSON serializer for datetime objects.

    Code is based on the answer of Jay Taylor in
    http://stackoverflow.com/questions/11875770/how-to-overcome-datetime-datetime-not-json-serializable-in-python

    Args:
        dt (datetime): The ``datetime`` instance.

    Returns:
        int: The int value of milliseconds since epoch.
    """
    if isinstance(dt, datetime) and dt.utcoffset() is not None:
        dt = dt - dt.utcoffset()
    millis = int(calendar.timegm(dt.timetuple()) * 1000 + dt.microsecond / 1000)
    return millis


def millis_to_datetime(millis: int) -> datetime:
    """Calculate the datetime from the given milliseconds value.

    Args:
        millis (int): An int value showing the millis from unix EPOCH

    Returns:
        datetime: The corresponding ``datetime`` instance to
            the given milliseconds.
    """
    epoch = datetime(1970, 1, 1, tzinfo=pytz.utc)
    return epoch + timedelta(milliseconds=millis)



# -*- coding: utf-8 -*-
"""Provides functionality to parse the version number from the VERSION file."""
import os
from typing import Union

VERSION: Union[None, str] = None
VERSION_FILE: str = os.path.join(os.path.dirname(__file__), "VERSION")
if os.path.isfile(VERSION_FILE):
    with open(VERSION_FILE, "r") as f:
        VERSION = f.read().strip()
__version__ = VERSION or "0.0.0"
"""str: The version of the package."""



# -*- coding: utf-8 -*-
"""Authentication related classes and functions situated here."""
import base64
import copy
import json
import os
import re
from datetime import datetime, timedelta
from typing import Any, Dict, List, Optional, TYPE_CHECKING, Union

import pytz

from sqlalchemy import Column, Enum, ForeignKey, Integer, String, Table
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.orm import Mapped, mapped_column, relationship, synonym, validates
from sqlalchemy.schema import UniqueConstraint

from stalker import defaults, log
from stalker.db.declarative import Base
from stalker.db.types import GenericDateTime
from stalker.models.entity import Entity, SimpleEntity
from stalker.models.mixins import ACLMixin
from stalker.models.status import Status
from stalker.utils import datetime_to_millis, millis_to_datetime

if TYPE_CHECKING:  # pragma: no cover
    from stalker.models.client import Client, ClientUser
    from stalker.models.department import Department, DepartmentUser
    from stalker.models.project import Project, ProjectUser
    from stalker.models.task import Task, TimeLog
    from stalker.models.ticket import Ticket
    from stalker.models.studio import Vacation

logger = log.get_logger(__name__)

LOGIN = "login"
LOGOUT = "logout"


class Permission(Base):
    """A class to hold permissions.

    Permissions in Stalker defines what one can do or do not. A Permission
    instance is composed by three attributes; access, action and class_name.

    Permissions for all the classes in SOM are generally created by Stalker
    when initializing the database.

    If you created any custom classes to extend SOM you are also responsible to
    create the Permissions for it by calling :meth:`stalker.db.register` and
    passing your class to it. See the :mod:`stalker.db` documentation for
    details.

    Example: Let say that you want to create a Permission specifying a Group of
    Users are allowed to create Projects::

    .. code-block:: Python

        from stalker import db
        from stalker import db
        from stalker.models.auth import User, Group, Permission

        # first setup the db with the default database
        #
        # stalker.db.init() will create all the Actions possible with the
        # SOM classes automatically
        #
        # What is left to you is to create the permissions
        db.setup()

        user1 = User(
            name='Test User',
            login='test_user1',
            password='1234',
            email='testuser1@test.com'
        )
        user2 = User(
            name='Test User',
            login='test_user2',
            password='1234',
            email='testuser2@test.com'
        )

        group1 = Group(name='users')
        group1.users = [user1, user2]

        # get the permissions for the Project class
        project_permissions = Permission.query\
            .filter(Permission.access='Allow')\
            .filter(Permission.action='Create')\
            .filter(Permission.class_name='Project')\
            .first()

        # now we have the permission specifying the allowance of creating a
        # Project

        # to make group1 users able to create a Project we simply add this
        # Permission to the groups permission attribute
        group1.permissions.append(permission)

        # and persist this information in the database
        DBSession.add(group)
        DBSession.commit()

    Args:
        access (str): An Enum value which can have the one of the values of
            ``Allow`` or ``Deny``.
        action (str): An Enum value from the list ['Create', 'Read', 'Update',
            'Delete', 'List']. Cannot be None. The list can be changed from
            stalker.config.Config.default_actions.

        class_name (str): The name of the class that this action is applied
            to. Cannot be None or an empty string.
    """

    __tablename__ = "Permissions"
    __table_args__ = (
        UniqueConstraint("access", "action", "class_name"),
        {"extend_existing": True},
    )

    id: Mapped[int] = mapped_column(primary_key=True)
    _access: Mapped[Optional[str]] = mapped_column(
        "access", Enum("Allow", "Deny", name="AccessNames")
    )
    _action: Mapped[Optional[str]] = mapped_column(
        "action", Enum(*defaults.actions, name="AuthenticationActions")
    )
    _class_name: Mapped[Optional[str]] = mapped_column("class_name", String(32))

    def __init__(self, access: str, action: str, class_name: str) -> None:
        super(Permission, self).__init__()
        self._access = self._validate_access(access)
        self._action = self._validate_action(action)
        self._class_name = self._validate_class_name(class_name)

    def __hash__(self) -> int:
        """Return the hash value of this instance.

        Because the __eq__ is overridden the __hash__ also needs to be overridden.

        Returns:
            int: The hash value.
        """
        return hash(self.access + self.action + self.class_name)

    def _validate_access(self, access: str) -> str:
        """Validate the given access value.

        Args:
            access (str): The access value to be validated.

        Raises:
            TypeError: If the given access value is not a str.
            ValueError: If the access is not a one of ["Access", "Deny"].

        Returns:
            str: The access value.
        """
        if not isinstance(access, str):
            raise TypeError(
                f"{self.__class__.__name__}.access should be an instance of str, "
                f"not {access.__class__.__name__}: '{access}'"
            )

        if access not in ["Allow", "Deny"]:
            raise ValueError(
                f'{self.__class__.__name__}.access should be "Allow" or "Deny" '
                f"not {access}"
            )

        return access

    def _access_getter(self) -> str:
        """Return the _access value.

        Returns:
            str: Returns the access value.
        """
        return self._access

    access: Mapped[Optional[str]] = synonym(
        "_access", descriptor=property(_access_getter)
    )

    def _validate_class_name(self, class_name: str) -> str:
        """Validate the given class_name value.

        Args:
            class_name (str): The class name.

        Raises:
            TypeError: If the class_name is not a str.

        Returns:
            str: The validated class_name.
        """
        if not isinstance(class_name, str):
            raise TypeError(
                f"{self.__class__.__name__}.class_name should be an instance of str, "
                f"not {class_name.__class__.__name__}: '{class_name}'"
            )

        return class_name

    def _class_name_getter(self) -> str:
        """Return the _class_name attribute value.

        Returns:
            str: The class name.
        """
        return self._class_name

    class_name: Mapped[str] = synonym(
        "_class_name", descriptor=property(_class_name_getter)
    )

    def _validate_action(self, action: str) -> str:
        """Validate the given action value.

        Args:
            action (str): The action value.

        Raises:
            TypeError: If the action is not a str value.
            ValueError: If the given action is not in the "defaults.actions" list.

        Returns:
            str: The validated action value.
        """
        if not isinstance(action, str):
            raise TypeError(
                f"{self.__class__.__name__}.action should be an instance of str, "
                f"not {action.__class__.__name__}: '{action}'"
            )

        if action not in defaults.actions:
            raise ValueError(
                f"{self.__class__.__name__}.action should be one of the values of "
                f"{defaults.actions} not '{action}'"
            )

        return action

    def _action_getter(self) -> str:
        """Return the _action value.

        Returns:
            str: Returns the action value.
        """
        return self._action

    action: Mapped[Optional[str]] = synonym(
        "_action", descriptor=property(_action_getter)
    )

    def __eq__(self, other: Any) -> bool:
        """Check if the other Permission is equal to this one.

        Args:
            other (Any): The other object.

        Returns:
            bool: True if the other object is a Permission instance and has the same
                access, action and class_name attributes.
        """
        return (
            isinstance(other, Permission)
            and other.access == self.access
            and other.action == self.action
            and other.class_name == self.class_name
        )


class Group(Entity, ACLMixin):
    """Creates groups for users to be used in authorization system.

    A Group instance is nothing more than a list of :class:`.User` s created
    to be able to assign permissions in a group level.

    The Group class, as with the :class:`.User` class, is mixed with the
    :class:`.ACLMixin` which adds ability to hold :class:`.Permission`
    instances and serve ACLs to Pyramid.

    Args:
        name (str): The name of this group.
        users list: A list of :class:`.User` instances holding the desired
            users in this group.
    """

    __auto_name__ = False
    __tablename__ = "Groups"
    __mapper_args__ = {"polymorphic_identity": "Group"}

    gid: Mapped[int] = mapped_column("id", ForeignKey("Entities.id"), primary_key=True)

    users: Mapped[Optional[List["User"]]] = relationship(
        "User",
        secondary="Group_Users",
        back_populates="groups",
        doc="""Users in this group.

        Accepts:class:`.User` instance.
        """,
    )

    def __init__(self, name="", users=None, permissions=None, **kwargs) -> None:
        if users is None:
            users = []

        if permissions is None:
            permissions = []

        kwargs.update({"name": name})
        super(Group, self).__init__(**kwargs)

        self.users = users
        self.permissions = permissions

    @validates("users")
    def _validate_users(self, key: str, user: "User") -> "User":
        """Validate the given user value.

        Args:
            key (str): The name of the validated column.
            user (User): The :class:`.User` instance.

        Raises:
            TypeError: If the given user is not a :class:`.User` instance.

        Returns:
            User: The validated :class:`.User` instance.
        """
        if not isinstance(user, User):
            raise TypeError(
                f"{self.__class__.__name__}.users should only contain "
                "instances of stalker.models.auth.User, "
                f"not {user.__class__.__name__}: '{user}'"
            )

        return user

    def __hash__(self) -> int:
        """Return the hash value of this instance.

        Because the __eq__ is overridden the __hash__ also needs to be overridden.

        Returns:
            int: The hash value.
        """
        return super(Group, self).__hash__()


class User(Entity, ACLMixin):
    """The user class is designed to hold data about a User in the system.

    .. note::
       .. versionadded 0.2.0: Task Watchers

       New to version 0.2.0 users can be assigned to a :class:`.Task` as a
       **Watcher**. Which can be used to inform the users in watchers list
       about the updates of certain Tasks.

    .. note::
       .. versionadded 0.2.0: Vacations

       It is now possible to define Vacations per user.

    .. note::
       .. versionadded 0.2.7: Resource Efficiency

    .. note::
       .. versionadded 0.2.11:

          Users not have a :attr:`.rate` attribute.

    Args:
        rate: For future usage a rate attribute is added to the User to record
            the daily cost of this user as a resource. It should be either 0 or
            a positive integer or float value. Default is 0.
        efficiency : The efficiency is a multiplier for a user as a resource to
            a task and defines how much of the time spent for that particular
            task is counted as an actual effort. The default value is 1.0,
            lowest possible value is 0.0 and there is no upper limit.

            The efficiency of a resource can be used for three purposes. First
            you can use it as a crude way to model a team. A team of 5 people
            should have an efficiency of 5.0. Keep in mind that you cannot track
            the members of the team individually if you use this feature. They
            always act as a group.

            Another use is to model performance variations between your resources.
            Again, this is a fairly crude mechanism and should be used with care.
            A resource that isn't every good at some task might be pretty good
            at another. This can't be taken into account as the resource efficiency
            can only set globally for all tasks.

            One another and interesting use is to model the availability of passive
            resources like a meeting room or a workstation or something that needs
            to be free for a task to take place but does not contribute to a task
            as an active resource.

            All resources that do not contribute effort to the task, that is a
            passive resource, should have an efficiency of 0.0. Again a typical
            example would be a conference room. It's necessary for a meeting,
            but it does not contribute any work.
        email (str): holds the e-mail of the user, should be in [part1]@[part2]
            format.
        login (str): This is the login name of the user, it should be all lower
            case. Giving a string that has uppercase letters, it will be converted
            to lower case. It cannot be an empty string or None and it cannot
            contain any white space inside.
        departments (List[Department]): It is the departments that the user is
            a part of. It should be a list of Department objects. One user can
            be listed in multiple departments.
        password (str): it is the password of the user, can contain any character.
            Stalker doesn't store the raw passwords of the users. To check a stored
            password with a raw password use :meth:`.check_password` and to set
            the password you can use the :attr:`.password` property directly.
        groups (List[Group]): It is a list of :class:`.Group` instances that this
            user belongs to.
        tasks (List[Task]): it is a list of Task objects which holds the tasks
            that this user has been assigned to.
        last_login (datetime): it is a datetime object holds the last login
            date of the user (not implemented yet).
    """

    __auto_name__ = False
    __tablename__ = "Users"
    __mapper_args__ = {"polymorphic_identity": "User"}

    user_id: Mapped[int] = mapped_column(
        "id", ForeignKey("Entities.id"), primary_key=True
    )

    departments = association_proxy(
        "department_role", "department", creator=lambda d: create_department_user(d)
    )

    department_role: Mapped[Optional[List["DepartmentUser"]]] = relationship(
        back_populates="user",
        cascade="all, delete-orphan",
        primaryjoin="Users.c.id==Department_Users.c.uid",
        doc="""A list of :class:`.Department` s that
        this user is a part of""",
    )

    companies = association_proxy(
        "company_role", "client", creator=lambda n: create_client_user(n)
    )

    company_role: Mapped[Optional[List["ClientUser"]]] = relationship(
        back_populates="user",
        cascade="all, delete-orphan",
        primaryjoin="Users.c.id==Client_Users.c.uid",
        doc="""A list of :class:`.Client` s that this user is a part of.""",
    )

    email: Mapped[str] = mapped_column(
        String(256),
        unique=True,
        nullable=False,
        doc="email of the user, accepts string",
    )

    password: Mapped[str] = mapped_column(
        String(256),
        nullable=False,
        doc="""The password of the user.

        It is scrambled before it is stored.
        """,
    )

    login: Mapped[str] = mapped_column(
        String(256),
        nullable=False,
        unique=True,
        doc="""The login name of the user.

        Cannot be empty.
        """,
    )

    authentication_logs: Mapped[Optional[List["AuthenticationLog"]]] = relationship(
        primaryjoin="AuthenticationLogs.c.uid==Users.c.id",
        back_populates="user",
        cascade="all, delete-orphan",
        doc="""A list of :class:`.AuthenticationLog` instances which holds the
        login/logout info for this :class:`.User`.
        """,
    )

    groups: Mapped[Optional[List["Group"]]] = relationship(
        secondary="Group_Users",
        back_populates="users",
        doc="""Permission groups that this users is a member of.

        Accepts :class:`.Group` object.
        """,
    )

    projects = association_proxy(
        "project_role", "project", creator=lambda p: create_project_user(p)
    )

    project_role: Mapped[Optional[List["ProjectUser"]]] = relationship(
        back_populates="user",
        cascade="all, delete-orphan",
        primaryjoin="Users.c.id==Project_Users.c.user_id",
    )

    tasks: Mapped[Optional[List["Task"]]] = relationship(
        secondary="Task_Resources",
        back_populates="resources",
        doc=""":class:`.Task` s assigned to this user.

        It is a list of :class:`.Task` instances.
        """,
    )

    watching: Mapped[Optional[List["Task"]]] = relationship(
        secondary="Task_Watchers",
        back_populates="watchers",
        doc=""":class:`.Tasks` s that this user is
        assigned as a watcher.

        It is a list of :class:`.Task` instances.
        """,
    )

    responsible_of: Mapped[Optional[List["Task"]]] = relationship(
        secondary="Task_Responsible",
        primaryjoin="Users.c.id==Task_Responsible.c.responsible_id",
        secondaryjoin="Task_Responsible.c.task_id==Tasks.c.id",
        back_populates="_responsible",
        doc="""A list of :class:`.Task` instances that this user is responsible
        of.""",
    )

    time_logs: Mapped[Optional[List["TimeLog"]]] = relationship(
        primaryjoin="TimeLogs.c.resource_id==Users.c.id",
        back_populates="resource",
        cascade="all, delete-orphan",
        doc="""A list of :class:`.TimeLog` instances which
        holds the time logs created for this :class:`.User`.
        """,
    )

    vacations: Mapped[Optional[List["Vacation"]]] = relationship(
        primaryjoin="Vacations.c.user_id==Users.c.id",
        back_populates="user",
        cascade="all, delete-orphan",
        doc="""A list of :class:`.Vacation` instances
        which holds the vacations created for this :class:`.User`
        """,
    )

    efficiency: Mapped[Optional[float]] = mapped_column(default=1.0)

    rate: Mapped[Optional[float]] = mapped_column(default=0.0)

    def __init__(
        self,
        name: Optional[str] = None,
        login: Optional[str] = None,
        email: Optional[str] = None,
        password: Optional[str] = None,
        departments: Optional["Department"] = None,
        companies: Optional["Client"] = None,
        groups: Optional["Group"] = None,
        efficiency: float = 1.0,
        rate: float = 0.0,
        **kwargs: Optional[Dict[str, Any]],
    ) -> None:
        kwargs["name"] = name

        super(User, self).__init__(**kwargs)

        self.login = login

        if departments is None:
            departments = []

        self.departments = departments

        if companies is None:
            companies = []
        self.companies = companies

        self.email = email

        # to be able to mangle the password do it like this
        self.password = password

        if groups is None:
            groups = []
        self.groups = groups

        self.tasks = []

        self.efficiency = efficiency
        self.rate = rate

    def __repr__(self) -> str:
        """Return the representation of the current User.

        Returns:
            str: The str representation of this User.
        """
        return f"<{self.name} ('{self.login}') (User)>"

    def __eq__(self, other: Any) -> bool:
        """Check if the other User is equal to this one.

        Args:
            other (Any): The other user instance.

        Returns:
            bool: If the other object is equal to this one, meaning that it is a User
                instance, has the same name, login and email values then returns True.
        """
        return (
            super(User, self).__eq__(other)
            and isinstance(other, User)
            and self.email == other.email
            and self.login == other.login
            and self.name == other.name
        )

    def __hash__(self) -> int:
        """Return the hash value of this instance.

        Because the __eq__ is overridden the __hash__ also needs to be overridden.

        Returns:
            int: The hash value.
        """
        return super(User, self).__hash__()

    @validates("login")
    def _validate_login(self, key: str, login: str) -> str:
        """Validate and format the given login value.

        Args:
            key (str): The name of the validated column.
            login (str): The login to be validated.

        Raises:
            TypeError: If the login is not a str.
            ValueError: If the login is an empty string after formatting.

        Returns:
            str: The validated and formatted login value.
        """
        if login is None:
            raise TypeError(f"{self.__class__.__name__}.login cannot be None")

        login = self._format_login(login)

        # raise a ValueError if the login is an empty string after formatting
        if login == "":
            raise ValueError(
                f"{self.__class__.__name__}.login cannot be an empty string"
            )

        logger.debug(f"name out: {login}")

        return login

    @validates("email")
    def _validate_email(self, key: str, email: str) -> str:
        """Validate the given email value.

        Args:
            key (str): The name of the validated column.
            email (str): The email to be validated.

        Raises:
            TypeError: If the given email is not a str.

        Returns:
            str: The validated email value.
        """
        # check if email is an instance of string
        if not isinstance(email, str):
            raise TypeError(
                f"{self.__class__.__name__}.email should be an instance of str, not "
                f"{email.__class__.__name__}: '{email}'"
            )
        return self._validate_email_format(email)

    def _validate_email_format(self, email: str) -> str:
        """Validate the email format.

        Args:
            email (str): The email value to be validated.

        Raises:
            ValueError: If the email doesn't have a "@" sign in it, or it has more than
                one "@" sign in it, or after formatting the account name part or the
                domain name part becomes an empty string.

        Returns:
            str: The validated email value.
        """
        # split the mail from @ sign
        splits = email.split("@")
        len_splits = len(splits)

        # there should be one and only one @ sign
        if len_splits > 2:
            raise ValueError(
                f"check the formatting of {self.__class__.__name__}.email, "
                "there are more than one @ sign"
            )

        if len_splits < 2:
            raise ValueError(
                f"check the formatting of {self.__class__.__name__}.email, "
                "there is no @ sign"
            )

        if splits[0] == "":
            raise ValueError(
                f"check the formatting of {self.__class__.__name__}.email, "
                "the name part is missing"
            )

        if splits[1] == "":
            raise ValueError(
                f"check the formatting {self.__class__.__name__}.email, "
                "the domain part is missing"
            )

        return email

    @classmethod
    def _format_login(cls, login: str) -> str:
        """Format the given login value.

        Args:
            login (str): The login value.

        Returns:
            str: The formatted login value.
        """
        # strip white spaces from start and end
        login = login.strip()

        # remove all the spaces
        login = login.replace(" ", "")

        # make it lowercase
        login = login.lower()

        # remove any illegal characters
        login = re.sub("[^\\(a-zA-Z0-9)]+", "", login)

        # remove any number at the beginning
        login = re.sub("^[0-9]+", "", login)

        return login

    @validates("password")
    def _validate_password(self, key: str, password: str) -> str:
        """Validate the given password value.

        Note:
            This function was updated to support both Python 2.7 and 3.5+. It will now
            explicitly convert the base64 bytes object into a string object.

        Args:
            key (str): The name of the validated column.
            password (str): The password value.

        Raises:
            TypeError: If the given password is None.
            ValueError: If the given password is an empty string.

        Returns:
            str: The mangled password.
        """
        if password is None:
            raise TypeError(f"{self.__class__.__name__}.password cannot be None")

        if password == "":
            raise ValueError(
                f"{self.__class__.__name__}.password cannot be an empty string"
            )

        # mangle the password
        mangled_password_bytes = base64.b64encode(password.encode("utf-8"))
        mangled_password_str = str(mangled_password_bytes.decode("utf-8"))
        return mangled_password_str

    def check_password(self, raw_password: str) -> bool:
        """Check the given raw_password.

        Check the given raw_password with the current User object's mangled password.
        Handles the encryption process behind the scene.

        Note:
            This function was updated to support both Python 2.7 and 3.5+.
            It will now compare the string (str) versions of the given
            raw_password and the current Users object encrypted password.

        Args:
            raw_password (str): The raw password.

        Returns:
            bool: If the given raw password matches the password stored in the db.
        """
        mangled_password_str = str(self.password)
        raw_password_bytes = base64.b64encode(bytes(raw_password.encode("utf-8")))
        raw_password_encrypted_str = str(raw_password_bytes.decode("utf-8"))
        return mangled_password_str == raw_password_encrypted_str

    @validates("groups")
    def _validate_groups(self, key: str, group: Group) -> Group:
        """Validate the given group value.

        Args:
            key (str): The name of the validated column.
            group (Group): The :class:`.Group` instance to be validated.

        Raises:
            TypeError: If the given group arg value is not a :class:`.Group` instance.

        Returns:
            Group: The validated :class:`.Group` instance.
        """
        if not isinstance(group, Group):
            raise TypeError(
                f"Any group in {self.__class__.__name__}.groups should be an instance "
                "of stalker.models.auth.Group, "
                f"not {group.__class__.__name__}: '{group}'"
            )

        return group

    @validates("tasks")
    def _validate_tasks(self, key: str, task: "Task") -> "Task":
        """Validate the given tasks attribute.

        Args:
            key (str): The name of the validated column.
            task (stalker.models.task.Task): The :class:`stalker.models.task.Task`
                instance to be validated.

        Raises:
            TypeError: If the given task arg value is not a
                :class:`stalker.models.task.Task` instance.

        Returns:
            Task: The validated :class:`stalker.models.task.Task` instance.
        """
        from stalker.models.task import Task

        if not isinstance(task, Task):
            raise TypeError(
                f"Any element in {self.__class__.__name__}.tasks should be an instance "
                f"of stalker.models.task.Task, not {task.__class__.__name__}: '{task}'"
            )
        return task

    @validates("watching")
    def _validate_watching(self, key: str, task: "Task") -> "Task":
        """Validate the given watching attribute.

        Args:
            key (str): The name of the validated column.
            task (stalker.models.task.Task): The :class:`stalker.models.task.Task`
                instance that the user will watch.

        Raises:
            TypeError: If the given task arg value is not a
                :class:`stalker.models.task.Task` instance.

        Returns:
            Task: The validated :class:`stalker.models.task.Task` instance.
        """
        from stalker.models.task import Task

        if not isinstance(task, Task):
            raise TypeError(
                f"Any element in {self.__class__.__name__}.watching should be an "
                "instance of stalker.models.task.Task, "
                f"not {task.__class__.__name__}: '{task}'"
            )
        return task

    @validates("vacations")
    def _validate_vacations(self, key: str, vacation: "Vacation") -> "Vacation":
        """Validate the given vacation value.

        Args:
            key (str): The name of the validated column.
            vacation (stalker.models.studio.Vacation): The
                :class:`stalker.models.studio.Vacation` instance.

        Raises:
            TypeError: If the given vacation argument value is not a
                :class:`stalker.models.vacation.Vacation` instance.

        Returns:
            Vacation: The validated vacation value.
        """
        from stalker.models.studio import Vacation

        if not isinstance(vacation, Vacation):
            raise TypeError(
                f"All of the elements in {self.__class__.__name__}.vacations should be "
                "a stalker.models.studio.Vacation instance, "
                f"not {vacation.__class__.__name__}: '{vacation}'"
            )
        return vacation

    @validates("efficiency")
    def _validate_efficiency(
        self, key: str, efficiency: Union[None, int, float]
    ) -> float:
        """Validate the given efficiency value.

        Args:
            key (str): The name of the validated column.
            efficiency (Union[None, int, float]): The efficiency of this User instance.
                This shows how efficient the user works. It is a number between 0-1. If
                None given, a default value of 1.0 will be used.

        Raises:
            TypeError: If the given efficiency is not one of [None, int, float].
            ValueError: If the given efficiency is a negative value.

        Returns:
            float: The validated efficiency value.
        """
        if efficiency is None:
            efficiency = 1.0

        if not isinstance(efficiency, (int, float)):
            raise TypeError(
                f"{self.__class__.__name__}.efficiency should be a float number "
                "greater or equal to 0.0, "
                f"not {efficiency.__class__.__name__}: '{efficiency}'"
            )

        if efficiency < 0:
            raise ValueError(
                f"{self.__class__.__name__}.efficiency should be a float number "
                f"greater or equal to 0.0, not {efficiency}"
            )

        return float(efficiency)

    @validates("rate")
    def _validate_rate(self, key: str, rate: Union[int, float]) -> float:
        """Validate the given rate value.

        Args:
            key (str): The name of the validated column.
            rate (Union[int, float]): An int or float value representing the User hourly
                rate.

        Raises:
            TypeError: If the given rate is not an int or float.
            ValueError: If the given rate is a negative number.

        Returns:
            float: The validated rate value.
        """
        if rate is None:
            rate = 0.0

        if not isinstance(rate, (int, float)):
            raise TypeError(
                f"{self.__class__.__name__}.rate should be a float number greater or "
                f"equal to 0.0, not {rate.__class__.__name__}: '{rate}'"
            )

        if rate < 0:
            raise ValueError(
                f"{self.__class__.__name__}.rate should be a float number greater or "
                f"equal to 0.0, not {rate}"
            )

        return float(rate)

    @property
    def tickets(self) -> List["Ticket"]:
        """Return the list of :class:`.Ticket` s that this user has.

        Returns:
            List[Ticket]: The list of :class:`.Ticket` instances which this user is the
                owner of.
        """
        # do it with sqlalchemy
        from stalker.models.ticket import Ticket

        return Ticket.query.filter(Ticket.owner == self).all()

    @property
    def open_tickets(self) -> List["Ticket"]:
        """Return the list of open :class:`.Ticket` s that this user has.

        Returns:
             List[Ticket]: A list of :class:`.Ticket` instances which are not closed and
                this user is assigned as the owner.
        """
        from stalker import Ticket

        return (
            Ticket.query.join(Status, Ticket.status)
            .filter(Ticket.owner == self)
            .filter(Status.code != "CLS")
            .all()
        )

    @property
    def to_tjp(self) -> str:
        """Return a TaskJuggler compatible str representation of this User instance.

        Returns:
            str: The TaskJuggler compatible representation of this User instance.
        """
        tab = "    "
        indent = tab
        tjp = f'resource {self.tjp_id} "{self.tjp_id}" {{'
        tjp += f"\n{indent}efficiency {self.efficiency}"
        for vacation in self.vacations:
            tjp += "\n"
            tjp += "\n".join(f"{indent}{line}" for line in vacation.to_tjp.split("\n"))
        tjp += "\n}"
        return tjp


class LocalSession(object):
    """A simple temporary session object which simple stores session data.

    This class will later be removed, it is here because we need a login window
    for the Qt user interfaces.

    On initialize it will load the SessionData from the users .strc folder
    """

    def __init__(self) -> None:
        self.logged_in_user_id = None
        self.valid_to = None
        self.session_data = None
        self.load()

    def load(self) -> None:
        """Load the data from the saved local session."""
        try:
            with open(LocalSession.session_file_full_path(), "r") as s:
                # try:
                json_object = json.load(s)
                valid_to = millis_to_datetime(json_object.get("valid_to"))
                if valid_to > datetime.now(pytz.utc):
                    # fill __dict__ with the loaded one
                    self.valid_to = valid_to
                    self.logged_in_user_id = json_object.get("logged_in_user_id")
        except IOError:
            pass

    @property
    def logged_in_user(self) -> "User":
        """Return the logged-in user.

        Returns:
            User: The logged-in user.
        """
        return User.query.filter_by(id=self.logged_in_user_id).first()

    def store_user(self, user: "User") -> None:
        """Store the given user instance.

        Args:
            user (User): The :class:`.User` instance.
        """
        if user:
            self.logged_in_user_id = user.id

    def save(self) -> None:
        """Remember the data in user local file system."""
        self.valid_to = datetime.now(pytz.utc) + timedelta(days=10)
        # serialize self
        dumped_data = json.dumps(
            {
                "valid_to": datetime_to_millis(self.valid_to),
                "logged_in_user_id": self.logged_in_user_id,
            },
        )
        logger.debug(f"dumped session data : {dumped_data}")
        self._write_data(dumped_data)

    def delete(self) -> None:
        """Remove the cache file."""
        try:
            os.remove(self.session_file_full_path())
        except OSError:
            pass

    @classmethod
    def session_file_full_path(cls) -> str:
        """Return the session file full path.

        Returns:
            str: The session file full path.
        """
        return os.path.normpath(
            os.path.join(
                defaults.local_storage_path, defaults.local_session_data_file_name
            )
        )

    def _write_data(self, data: str) -> None:
        """Write the given data to the local session file.

        Args:
            data (str): The data to be written (generally serialized LocalSession class
                itself)
        """
        file_full_path = self.session_file_full_path()

        # create the path first
        file_path = os.path.dirname(file_full_path)
        try:
            os.makedirs(file_path)
        except OSError:
            # dir exists
            pass
        finally:
            with open(file_full_path, "w") as data_file:
                data_file.write(data)


class Role(Entity):
    """Defines a User role.

    .. versionadded 0.2.11: Roles

    When :class:`.User` s are assigned to a
    :class:`.Client`/:class:`.Department`, they also can be assigned to a role
    for that client/department.

    Also, because Users can be assigned to multiple clients/departments they
    can have different roles for each of this clients/departments.

    The duty of this class is to defined different roles that can be reused
    when required. So one can defined a **Lead** role and then assign a User to
    a department with its role is set to "lead". This essentially generalizes
    the previous implementation of now removed *Department.lead* attribute.
    """

    __auto_name__ = False
    __tablename__ = "Roles"
    __mapper_args__ = {"polymorphic_identity": "Role"}

    role_id: Mapped[int] = mapped_column(
        "id", ForeignKey("Entities.id"), primary_key=True
    )

    def __init__(self, **kwargs) -> None:
        super(Role, self).__init__(**kwargs)


def create_department_user(department: "Department") -> "DepartmentUser":
    """Create DepartmentUser instance on association proxy.

    Args:
        department (stalker.models.department.Department): The
            :class:`stalker.models.department.Department` instance.

    Returns:
        stalker.models.department.DepartmentUser: The
            :class:`stalker.models.department.DepartmentUser` instance.
    """
    from stalker.models.department import DepartmentUser

    return DepartmentUser(department=department)


def create_client_user(client: "Client") -> "ClientUser":
    """Create ClientUser instance on association proxy.

    Args:
        client (stalker.models.client.Client): The
            :class:`stalker.models.project.Project` instance.

    Returns:
        stalker.models.client.ClientUser: The
            :class:`stalker.models.client.ClientUser` instance.
    """
    from stalker.models.client import ClientUser

    return ClientUser(client=client)


def create_project_user(project: "Project") -> "ProjectUser":
    """Create ProjectUser instance on association proxy.

    Args:
        project (stalker.models.project.Project): The
            :class:`stalker.models.project.Project` instance.

    Returns:
        stalker.models.project.ProjectUser: The
            :class:`stalker.models.project.ProjectUser` instance.
    """
    from stalker.models.project import ProjectUser

    return ProjectUser(project=project)


# Group_Users
Group_Users = Table(
    "Group_Users",
    Base.metadata,
    Column("uid", Integer, ForeignKey("Users.id"), primary_key=True),
    Column("gid", Integer, ForeignKey("Groups.id"), primary_key=True),
)


class AuthenticationLog(SimpleEntity):
    """Keeps track of login/logout dates and the action (login or logout)."""

    __auto_name__ = True
    __tablename__ = "AuthenticationLogs"
    __mapper_args__ = {"polymorphic_identity": "AuthenticationLog"}

    log_id: Mapped[int] = mapped_column(
        "id", ForeignKey("SimpleEntities.id"), primary_key=True
    )
    user_id: Mapped[int] = mapped_column("uid", Integer, ForeignKey("Users.id"))
    user: Mapped[User] = relationship(
        primaryjoin="AuthenticationLogs.c.uid==Users.c.id",
        uselist=False,
        back_populates="authentication_logs",
        doc="The :class:`.User` instance that this AuthenticationLog is created for",
    )
    action: Mapped[str] = mapped_column(Enum(LOGIN, LOGOUT, name="ActionNames"))
    date: Mapped[datetime] = mapped_column(GenericDateTime)

    def __init__(self, user=None, date=None, action=LOGIN, **kwargs) -> None:
        super(AuthenticationLog, self).__init__(**kwargs)
        self.user = user
        self.date = date
        self.action = action

    def __lt__(self, other: "AuthenticationLog") -> bool:
        """Make this object order-able.

        Args:
            other (.AuthenticationLog): The other :class:`.AuthenticationLog`
                instance.

        Returns:
            Tuple(str, str): The str key to be used for ordering.
        """
        return self.date < other.date

    @validates("user")
    def __validate_user__(self, key: str, user: "User") -> "User":
        """Validate the given user argument value.

        Args:
            key (str): The name of the validated column.
            user (User): The :class:`.User` instance to be validated.

        Raises:
            TypeError: If the given user args is not a :class:`.User` instance.

        Returns:
            .User: The validated :class:`.User` instance.
        """
        if not isinstance(user, User):
            raise TypeError(
                f"{self.__class__.__name__}.user should be a User instance, "
                f"not {user.__class__.__name__}: '{user}'"
            )

        return user

    @validates("action")
    def __validate_action__(self, key: str, action: str) -> str:
        """Validate the given action argument value.

        Args:
            key (str): The name of the validated column.
            action (str): One of LOGIN or LOGOUT enum values.

        Raises:
            ValueError: If the given value is not one of LOGIN or LOGOUT.

        Returns:
            str: The validated action value.
        """
        if action is None:
            action = copy.copy(LOGIN)

        if action not in [LOGIN, LOGOUT]:
            raise ValueError(
                f'{self.__class__.__name__}.action should be one of "login" or '
                f'"logout", not "{action}"'
            )

        return action

    @validates("date")
    def __validate_date__(self, key: str, date: datetime) -> datetime:
        """Validate the given date value.

        Args:
            key (str): The name of the validated column.
            date (datetime): The datetime instance.

        Raises:
            TypeError: If the given date is not a datetime instance.

        Returns:
            datetime: Returns the validated datetime instance.
        """
        if date is None:
            date = datetime.now(pytz.utc)

        if not isinstance(date, datetime):
            raise TypeError(
                f"{self.__class__.__name__}.date should be a datetime.datetime "
                f"instance, not {date.__class__.__name__}: '{date}'"
            )

        return date



# -*- coding: utf-8 -*-
"""Stalker specific data types are situated here."""
import datetime
import json
from typing import Any, Dict, TYPE_CHECKING, Union

import pytz

from sqlalchemy.types import DateTime, JSON, TEXT, TypeDecorator

import tzlocal

if TYPE_CHECKING:  # pragma: no cover
    from sqlalchemy.engine.interfaces import Dialect


class JSONEncodedDict(TypeDecorator):
    """Stores and retrieves JSON as TEXT."""

    impl = TEXT

    def process_bind_param(self, value: Union[None, Any], dialect: "Dialect") -> str:
        """Process bind param.

        Args:
            value (Union[None, Any]): The object to convert to JSON.
            dialect (sqlalchemy.engine.interface.Dialect): The dialect.

        Returns:
            str: The str representation of the JSON data.
        """
        if value is not None:
            value = json.dumps(value)
        return value

    def process_result_value(
        self, value: Union[None, str], dialect: "Dialect"
    ) -> Union[None, Dict[str, Any]]:
        """Process result value.

        Args:
            value (Union[None, Any]): The str representation of the JSON data.
            dialect (sqlalchemy.engine.interface.Dialect): The dialect.

        Returns:
            dict: The dict representation of the JSON data.
        """
        return_value: Union[None, Dict[str, Any]] = None
        if value is not None:
            return_value = json.loads(value)
        return return_value


GenericJSON = JSON().with_variant(JSONEncodedDict, "sqlite")
"""A JSON variant that can be used both for PostgreSQL and SQLite3

It will be native JSON for PostgreSQL and JSONEncodedDict for SQLite3
"""


class DateTimeUTC(TypeDecorator):
    """Store UTC internally without the timezone info.

    Inject timezone info as the data comes back from db.
    """

    impl = DateTime

    def process_bind_param(self, value: Any, dialect: str) -> datetime.datetime:
        """Process bind param.

        Args:
            value (Any): The value.
            dialect (str): The dialect.

        Returns:
            datetime.datetime: The datetime value with UTC timezone.
        """
        if value is not None:
            # convert the datetime object to have UTC
            # and strip the datetime value out (which is automatic for SQLite3)
            value = value.astimezone(pytz.utc)
        return value

    def process_result_value(self, value: Any, dialect: str) -> datetime.datetime:
        """Process result value.

        Args:
            value (Any): The value.
            dialect (str): The dialect.

        Returns:
            datetime.datetime: The datetime value with UTC timezone.
        """
        if value is not None:
            # inject utc and then convert to local timezone
            local_tz = tzlocal.get_localzone()
            value = value.replace(tzinfo=pytz.utc).astimezone(local_tz)
        return value


GenericDateTime = DateTime(timezone=True).with_variant(DateTimeUTC, "sqlite")
"""A DateTime variant that can be used with both PostgreSQL and SQLite3 and
adds support to timezones in SQLite3.
"""



# -*- coding: utf-8 -*-
"""Errors for the system.

This module contains the Errors in Stalker.
"""


class LoginError(Exception):
    """Raised when the login information is not correct."""

    def __init__(self, value="") -> None:
        super(LoginError, self).__init__(value)
        self.value = value

    def __str__(self) -> str:
        """Return the string representation of this exception.

        Returns:
            str: The string representation of this exception.
        """
        return self.value


class CircularDependencyError(Exception):
    """Raised when there is circular dependencies within Tasks."""

    def __init__(self, value="") -> None:
        super(CircularDependencyError, self).__init__(value)
        self.value = value

    def __str__(self) -> str:
        """Return the string representation of this exception.

        Returns:
            str: The string representation of this exception.
        """
        return self.value


class OverBookedError(Exception):
    """Raised when a resource is booked more than once for the same time period."""

    def __init__(self, value="") -> None:
        super(OverBookedError, self).__init__(value)
        self.value = value

    def __str__(self) -> str:
        """Return the string representation of this exception.

        Returns:
            str: The string representation of this exception.
        """
        return self.value


class StatusError(Exception):
    """Raised when the status of an entity is not suitable for the desired action."""

    def __init__(self, value="") -> None:
        super(StatusError, self).__init__(value)
        self.value = value

    def __str__(self) -> str:
        """Return the string representation of this exception.

        Returns:
            str: The string representation of this exception.
        """
        return self.value


class DependencyViolationError(Exception):
    """Raised when a TimeLog violates the dependency relation between tasks."""

    def __init__(self, value="") -> None:
        super(DependencyViolationError, self).__init__(value)
        self.value = value

    def __str__(self) -> str:
        """Return the string representation of this exception.

        Returns:
            str: The string representation of this exception.
        """
        return self.value



# -*- coding: utf-8 -*-
"""Enum classes are situated here."""

from enum import Enum, IntEnum
from typing import Union

from sqlalchemy import Enum as saEnum, Integer, TypeDecorator


class ScheduleConstraint(IntEnum):
    """The schedule constraint enum."""

    NONE = 0
    Start = 1
    End = 2
    Both = 3

    def __repr__(self) -> str:
        """Return the enum name for str().

        Returns:
            str: The name as the string representation of this
                ScheduleConstraint.
        """
        return self.name if self.name != "NONE" else "None"

    __str__ = __repr__

    @classmethod
    def to_constraint(
        cls, constraint: Union[int, str, "ScheduleConstraint"]
    ) -> "ScheduleConstraint":
        """Validate and return type enum from an input int or str value.

        Args:
            constraint (Union[str, ScheduleConstraint]): Input `constraint` value.

        Raises:
            TypeError: Input value type is invalid.
            ValueError: Input value is invalid.

        Returns:
            ScheduleConstraint: ScheduleConstraint value.
        """
        # Check if it's a valid str type for a constraint.
        if constraint is None:
            constraint = ScheduleConstraint.NONE

        if not isinstance(constraint, (int, str, ScheduleConstraint)):
            raise TypeError(
                "constraint should be a ScheduleConstraint enum value or an "
                "int or a str, "
                f"not {constraint.__class__.__name__}: '{constraint}'"
            )

        if isinstance(constraint, str):
            constraint_name_lut = dict(
                [
                    (c.name.lower(), c.name.title() if c.name != "NONE" else "NONE")
                    for c in cls
                ]
            )
            # also add int values
            constraint_lower_case = constraint.lower()
            if constraint_lower_case not in constraint_name_lut:
                raise ValueError(
                    "constraint should be a ScheduleConstraint enum value or "
                    "one of {}, not '{}'".format(
                        [e.name.title() for e in cls], constraint
                    )
                )

            # Return the enum status for the status value.
            return cls.__members__[constraint_name_lut[constraint_lower_case]]
        else:
            return ScheduleConstraint(constraint)


class ScheduleConstraintDecorator(TypeDecorator):
    """Store ScheduleConstraint as an integer and restore as ScheduleConstraint."""

    cache_ok = True
    impl = Integer

    def process_bind_param(self, value, dialect) -> int:
        """Return the integer value of the ScheduleConstraint.

        Args:
            value (ScheduleConstraint): The ScheduleConstraint value.
            dialect (str): The name of the dialect.

        Returns:
            int: The value of the ScheduleConstraint.
        """
        # just return the value
        return value.value

    def process_result_value(self, value: int, dialect: str) -> ScheduleConstraint:
        """Return a ScheduleConstraint.

        Args:
            value (int): The integer value.
            dialect (str): The name of the dialect.

        Returns:
            ScheduleConstraint: ScheduleConstraint created from the DB data.
        """
        return ScheduleConstraint.to_constraint(value)


class TimeUnit(Enum):
    """The time unit enum."""

    Minute = "min"
    Hour = "h"
    Day = "d"
    Week = "w"
    Month = "m"
    Year = "y"

    def __str__(self) -> str:
        """Return the string representation.

        Returns:
            str: The string representation.
        """
        return str(self.value)

    @classmethod
    def to_unit(cls, unit: Union[str, "TimeUnit"]) -> "TimeUnit":
        """Convert the given unit value to a TimeUnit enum.

        Args:
            unit (Union[str, TimeUnit]): The value to convert to a
                TimeUnit.

        Raises:
            TypeError: Input value type is invalid.
            ValueError: Input value is invalid.

        Returns:
            TimeUnit: The enum.
        """
        if not isinstance(unit, (str, TimeUnit)):
            raise TypeError(
                "unit should be a TimeUnit enum value or one of {}, "
                "not {}: '{}'".format(
                    [u.name.title() for u in cls] + [u.value for u in cls],
                    unit.__class__.__name__,
                    unit,
                )
            )
        if isinstance(unit, str):
            unit_name_lut = dict([(u.name.lower(), u.name) for u in cls])
            unit_name_lut.update(dict([(u.value.lower(), u.name) for u in cls]))
            unit_lower_case = unit.lower()
            if unit_lower_case not in unit_name_lut:
                raise ValueError(
                    "unit should be a TimeUnit enum value or one of {}, "
                    "not '{}'".format(
                        [u.name.title() for u in cls] + [u.value for u in cls], unit
                    )
                )

            return cls.__members__[unit_name_lut[unit_lower_case]]

        return unit


class TimeUnitDecorator(TypeDecorator):
    """Store TimeUnit as an str and restore as TimeUnit."""

    cache_ok = True
    impl = saEnum(*[u.value for u in TimeUnit], name="TimeUnit")

    def process_bind_param(self, value: TimeUnit, dialect: str) -> str:
        """Return the str value of the TimeUnit.

        Args:
            value (TimeUnit): The TimeUnit value.
            dialect (str): The name of the dialect.

        Returns:
            str: The value of the TimeUnit.
        """
        # just return the value
        return value.value

    def process_result_value(self, value: str, dialect: str) -> TimeUnit:
        """Return a TimeUnit.

        Args:
            value (str): The string value to convert to TimeUnit.
            dialect (str): The name of the dialect.

        Returns:
            TimeUnit: The TimeUnit which is created from the DB data.
        """
        return TimeUnit.to_unit(value)


class ScheduleModel(Enum):
    """The schedule model enum."""

    Effort = "effort"
    Duration = "duration"
    Length = "length"

    def __str__(self) -> str:
        """Return the string representation.

        Returns:
            str: The string representation.
        """
        return str(self.value)

    @classmethod
    def to_model(cls, model: Union[str, "ScheduleModel"]) -> "ScheduleModel":
        """Convert the given model value to a ScheduleModel enum.

        Args:
            model (Union[str, ScheduleModel]): The value to convert to a
                ScheduleModel.

        Raises:
            TypeError: Input value type is invalid.
            ValueError: Input value is invalid.

        Returns:
            ScheduleModel: The enum.
        """
        if not isinstance(model, (str, ScheduleModel)):
            raise TypeError(
                "model should be a ScheduleModel enum value or one of {}, "
                "not {}: '{}'".format(
                    [m.name.title() for m in cls] + [m.value for m in cls],
                    model.__class__.__name__,
                    model,
                )
            )
        if isinstance(model, str):
            model_name_lut = dict([(m.name.lower(), m.name) for m in cls])
            model_name_lut.update(dict([(m.value.lower(), m.name) for m in cls]))
            model_lower_case = model.lower()
            if model_lower_case not in model_name_lut:
                raise ValueError(
                    "model should be a ScheduleModel enum value or one of {}, "
                    "not '{}'".format(
                        [m.name.title() for m in cls] + [m.value for m in cls], model
                    )
                )

            return cls.__members__[model_name_lut[model_lower_case]]

        return model


class ScheduleModelDecorator(TypeDecorator):
    """Store ScheduleModel as a str and restore as ScheduleModel."""

    cache_ok = True
    impl = saEnum(*[m.value for m in ScheduleModel], name="ScheduleModel")

    def process_bind_param(self, value, dialect) -> str:
        """Return the str value of the ScheduleModel.

        Args:
            value (ScheduleModel): The ScheduleModel value.
            dialect (str): The name of the dialect.

        Returns:
            str: The value of the ScheduleModel.
        """
        # just return the value
        return value.value

    def process_result_value(self, value: str, dialect: str) -> ScheduleModel:
        """Return a ScheduleModel.

        Args:
            value (str): The string value to convert to ScheduleModel.
            dialect (str): The name of the dialect.

        Returns:
            ScheduleModel: The ScheduleModel created from the DB data.
        """
        return ScheduleModel.to_model(value)


class DependencyTarget(Enum):
    """The dependency target enum."""

    OnStart = "onstart"
    OnEnd = "onend"

    def __str__(self) -> str:
        """Return the string representation.

        Returns:
            str: The string representation.
        """
        return str(self.value)

    @classmethod
    def to_target(cls, target: Union[str, "DependencyTarget"]) -> "DependencyTarget":
        """Convert the given target value to a DependencyTarget enum.

        Args:
            target (Union[str, DependencyTarget]): The value to convert to a
                DependencyTarget.

        Raises:
            TypeError: Input value type is invalid.
            ValueError: Input value is invalid.

        Returns:
            DependencyTarget: The enum.
        """
        if not isinstance(target, (str, DependencyTarget)):
            raise TypeError(
                "target should be a DependencyTarget enum value or one of {}, "
                "not {}: '{}'".format(
                    [t.name for t in cls] + [t.value for t in cls],
                    target.__class__.__name__,
                    target,
                )
            )
        if isinstance(target, str):
            target_name_lut = dict([(t.name.lower(), t.name) for t in cls])
            target_name_lut.update(dict([(t.value.lower(), t.name) for t in cls]))
            target_lower_case = target.lower()
            if target_lower_case not in target_name_lut:
                raise ValueError(
                    "target should be a DependencyTarget enum value or one of {}, "
                    "not '{}'".format(
                        [t.name for t in cls] + [t.value for t in cls], target
                    )
                )

            return cls.__members__[target_name_lut[target_lower_case]]

        return target


class DependencyTargetDecorator(TypeDecorator):
    """Store DependencyTarget as an enum and restore as DependencyTarget."""

    cache_ok = True
    impl = saEnum(*[m.value for m in DependencyTarget], name="TaskDependencyTarget")

    def process_bind_param(self, value, dialect) -> str:
        """Return the str value of the DependencyTarget.

        Args:
            value (DependencyTarget): The DependencyTarget value.
            dialect (str): The name of the dialect.

        Returns:
            str: The value of the DependencyTarget.
        """
        # just return the value
        return value.value

    def process_result_value(self, value: str, dialect: str) -> DependencyTarget:
        """Return a DependencyTarget.

        Args:
            value (str): The string value to convert to DependencyTarget.
            dialect (str): The name of the dialect.

        Returns:
            DependencyTarget: The DependencyTarget created from str.
        """
        return DependencyTarget.to_target(value)


class TraversalDirection(IntEnum):
    """The traversal direction enum."""

    DepthFirst = 0
    BreadthFirst = 1

    def __repr__(self) -> str:
        """Return the enum name for str().

        Returns:
            str: The name as the string representation of this
                ScheduleConstraint.
        """
        return self.name if self.name != "NONE" else "None"

    __str__ = __repr__

    @classmethod
    def to_direction(
        cls, direction: Union[int, str, "TraversalDirection"]
    ) -> "TraversalDirection":
        """Convert the given direction value to a TraversalDirection enum.

        Args:
            direction (Union[int, str, TraversalDirection]): The value to
                convert to a TraversalDirection.

        Raises:
            TypeError: Input value type is invalid.
            ValueError: Input value is invalid.

        Returns:
            TraversalDirection: The enum.
        """
        if not isinstance(direction, (int, str, TraversalDirection)):
            raise TypeError(
                "direction should be a TraversalDirection enum value "
                "or one of {}, not {}: '{}'".format(
                    [d.name for d in cls] + [d.value for d in cls],
                    direction.__class__.__name__,
                    direction,
                )
            )
        if isinstance(direction, str):
            direction_name_lut = dict([(d.name.lower(), d.name) for d in cls])
            direction_name_lut.update(dict([(d.value, d.name) for d in cls]))
            direction_lower_case = direction.lower()
            if direction_lower_case not in direction_name_lut:
                raise ValueError(
                    "direction should be a TraversalDirection enum value or "
                    "one of {}, not '{}'".format(
                        [d.name for d in cls] + [d.value for d in cls],
                        direction,
                    )
                )

            return cls.__members__[direction_name_lut[direction_lower_case]]

        return direction



# -*- coding: utf-8 -*-
"""Logging related functions are situated here.

This module allows registering any number of logger so that it is possible
to update the logging level all together at runtime (without relaying on weird
hacks).
"""

import logging

logging.basicConfig()
logging_level = logging.INFO
loggers = []


def get_logger(name: str) -> logging.Logger:
    """Get a logger.

    Args:
        name (str): The name of the logger.

    Returns:
        logging.Logger: The logger.
    """
    logger = logging.getLogger(name)
    register_logger(logger)
    return logger


def register_logger(logger: logging.Logger) -> None:
    """Register logger.

    Args:
        logger (logging.Logger): A logging.Logger instance.

    Raises:
        TypeError: If the logger is not a logging.Logger instance.
    """
    if not isinstance(logger, logging.Logger):
        raise TypeError(
            "logger should be a logging.Logger instance, not {}: '{}'".format(
                logger.__class__.__name__, logger
            )
        )

    if logger not in loggers:
        loggers.append(logger)

    logger.setLevel(logging_level)


def set_level(level: int) -> None:
    """Update all registered loggers to the given level.

    Args:
        level (int): The logging level. The value should be valid with the
            logging library and should be one of [NOTSET, DEBUG, INFO, WARN,
            WARNING, ERROR, FATAL, CRITICAL] of the logging library (or anything
            that is registered as a proper logging level).

    Raises:
        TypeError: If level is not an integer.
        ValueError: If level is not a valid value for the logging library.
    """
    if not isinstance(level, int):
        raise TypeError(
            "level should be an integer value one of [0, 10, 20, 30, 40, 50] "
            "or [NOTSET, DEBUG, INFO, WARN, WARNING, ERROR, FATAL, CRITICAL] "
            "of the logging library, not {}: '{}'".format(
                level.__class__.__name__, level
            )
        )

    level_names = logging._levelToName

    if level not in level_names:
        raise ValueError(
            "level should be an integer value one of [0, 10, 20, 30, 40, 50] "
            "or [NOTSET, DEBUG, INFO, WARN, WARNING, ERROR, FATAL, CRITICAL] "
            "of the logging library, not {}.".format(level)
        )

    for logger in loggers:
        logger.setLevel(level)



# -*- coding: utf-8 -*-
"""Image format related classes and utility functions are situated here."""
from typing import Any, Optional, Union

from sqlalchemy import ForeignKey, Integer
from sqlalchemy.orm import Mapped, mapped_column, validates

from stalker.log import get_logger
from stalker.models.entity import Entity

logger = get_logger(__name__)


class ImageFormat(Entity):
    """Common image formats for the :class:`.Project` s.

    Args:
        width (Union[int, float]): The width of the format, it cannot be zero or
            negative, if a float number is given it will be converted to integer.

        height (Union[int, float]): The height of the format, it cannot be zero or
            negative, if a float number is given it will be converted to integer.

        pixel_aspect ((Union[int, float])): The pixel aspect ratio of the current
            ImageFormat object, it cannot be zero or negative, and if given as an
            integer it will be converted to a float, the default value is 1.0.

        print_resolution (Union[int, float]): The print resolution of the ImageFormat
            given as DPI (dot-per-inch). It cannot be zero or negative.
    """

    __auto_name__ = False
    __tablename__ = "ImageFormats"
    __mapper_args__ = {"polymorphic_identity": "ImageFormat"}

    imageFormat_id: Mapped[int] = mapped_column(
        "id",
        Integer,
        ForeignKey("Entities.id"),
        primary_key=True,
    )

    width: Mapped[Optional[int]] = mapped_column(
        doc="""The width of this format.

        * the width should be set to a positive non-zero integer
        * integers are also accepted but will be converted to float
        * for improper inputs the object will raise an exception.
        """,
    )

    height: Mapped[Optional[int]] = mapped_column(
        doc="""The height of this format

        * the height should be set to a positive non-zero integer
        * integers are also accepted but will be converted to float
        * for improper inputs the object will raise an exception.
        """,
    )

    pixel_aspect: Mapped[Optional[float]] = mapped_column(
        default=1.0,
        doc="""The pixel aspect ratio of this format.

        * the pixel_aspect should be set to a positive non-zero float
        * integers are also accepted but will be converted to float
        * for improper inputs the object will raise an exception
        """,
    )

    print_resolution: Mapped[Optional[float]] = mapped_column(
        default=300.0,
        doc="""The print resolution of this format

        * it should be set to a positive non-zero float or integer
        * integers are also accepted but will be converted to float
        * for improper inputs the object will raise an exception.
        """,
    )

    def __init__(
        self,
        width: Union[int, float],
        height: Union[int, float],
        pixel_aspect: Optional[Union[int, float]] = 1.0,
        print_resolution: Optional[Union[int, float]] = 300,
        **kwargs,
    ) -> None:
        super(ImageFormat, self).__init__(**kwargs)

        self.width = width
        self.height = height
        self.pixel_aspect = pixel_aspect
        self.print_resolution = print_resolution
        # self._device_aspect = 1.0

    @validates("width")
    def _validate_width(self, key: str, width: Union[int, float]) -> int:
        """Validate the given width.

        Args:
            key (str): The name of the validated column.
            width (Union[int, float]): The width value to be validated.

        Raises:
            TypeError: If the width is not an int or float.
            ValueError: If the width is 0 or a negative value.

        Returns:
            int: The validated width value.
        """
        if not isinstance(width, (int, float)):
            raise TypeError(
                f"{self.__class__.__name__}.width should be an instance of int or "
                f"float, not {width.__class__.__name__}: '{width}'"
            )

        if width <= 0:
            raise ValueError(
                f"{self.__class__.__name__}.width cannot be zero or negative"
            )

        return int(width)

    @validates("height")
    def _validate_height(self, key: str, height: Union[int, float]) -> int:
        """Validate the given height.

        Args:
            key (str): The name of the validated column.
            height (Union[int, float]): The height value to be validated.

        Raises:
            TypeError: If the height is not an int or float.
            ValueError: If the height is 0 or a negative value.

        Returns:
            int: The validated height value.
        """
        if not isinstance(height, (int, float)):
            raise TypeError(
                f"{self.__class__.__name__}.height should be an instance of int or "
                f"float, not {height.__class__.__name__}: '{height}'"
            )

        if height <= 0:
            raise ValueError(
                f"{self.__class__.__name__}.height cannot be zero or negative"
            )

        return int(height)

    @validates("pixel_aspect")
    def _validate_pixel_aspect(
        self, key: str, pixel_aspect: Union[int, float]
    ) -> float:
        """Validate the given pixel aspect.

        Args:
            key (str): The name of the validated column.
            pixel_aspect (Union[int, float]): The pixel_aspect value to be validated.

        Raises:
            TypeError: If the given pixel_aspect value is not an int for float.
            ValueError: If the pixel_aspect is 0 or a negative value.

        Returns:
            float: The validated pixel_aspect value.
        """
        if not isinstance(pixel_aspect, (int, float)):
            raise TypeError(
                f"{self.__class__.__name__}.pixel_aspect should be an instance of int "
                f"or float, not {pixel_aspect.__class__.__name__}: '{pixel_aspect}'"
            )

        if pixel_aspect <= 0:
            raise ValueError(
                f"{self.__class__.__name__}.pixel_aspect cannot be zero or a negative "
                "value"
            )

        return float(pixel_aspect)

    @validates("print_resolution")
    def _validate_print_resolution(
        self, key: str, print_resolution: Union[int, float]
    ) -> float:
        """Validate the print resolution value.

        Args:
            key (str): The name of the validated column.
            print_resolution (Union[int, float]): The print_resolution value to be
                validated.

        Raises:
            TypeError: If the given print_resolution is not an int or float.
            ValueError: If the print_resolution is 0 or negative value.

        Returns:
            float: The validated print_resolution value.
        """
        if not isinstance(print_resolution, (int, float)):
            raise TypeError(
                f"{self.__class__.__name__}.print_resolution should be an instance of "
                "int or float, "
                f"not {print_resolution.__class__.__name__}: '{print_resolution}'"
            )

        if print_resolution <= 0:
            raise ValueError(
                f"{self.__class__.__name__}.print_resolution cannot be zero or negative"
            )

        return float(print_resolution)

    @property
    def device_aspect(self) -> float:
        """Return the device aspect.

        Because the device_aspect is calculated from the width/height*pixel
        formula, this property is read-only.

        Returns:
            float: The device aspect ratio.
        """
        return float(self.width) / float(self.height) * self.pixel_aspect

    def __eq__(self, other: Any) -> bool:
        """Check if other is equal to this ImageFormat.

        Args:
            other (Any): The object to check the equality of.

        Returns:
            bool: True if the other is an ImageFormat instance and the width, height and
                pixel_aspect values are all equal.
        """
        return (
            super(ImageFormat, self).__eq__(other)
            and isinstance(other, ImageFormat)
            and self.width == other.width
            and self.height == other.height
            and self.pixel_aspect == other.pixel_aspect
        )

    def __hash__(self) -> int:
        """Return the hash value of this instance.

        Because the __eq__ is overridden the __hash__ also needs to be overridden.

        Returns:
            int: The hash value.
        """
        return super(ImageFormat, self).__hash__()
